//  VirtualDub - Video processing and capture application
//  Copyright (C) 1998-2005 Avery Lee
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//  This version has been modified by fccHandler, September 2007

#include "stdafx.h"
#include <process.h>
#include <windows.h>
#include <commctrl.h>
#include <fcntl.h>
#include <io.h>

#include "AudioSource.h"
#include "VideoSource.h"
#include "FastReadStream.h"
#include <vd2/system/error.h>
#include <vd2/system/fraction.h>
#include <vd2/system/log.h>
#include <vd2/system/file.h>
#include <vd2/Dita/resources.h>

#include "misc.h"
#include "mpeg.h"
#include "DataVector.h"
#include "resource.h"
#include "gui.h"
//#include "projectui.h"
#include <vd2/system/cpuaccel.h>
#include <vd2/Priss/decoder.h>
#include <vd2/Meia2/MPEG2Decoder.h>
#include <vd2/Meia2/MPEGPredict.h>
#include <vd2/Meia2/MPEGIDCT.h>
#include <vd2/Meia2/MPEGConvert.h>

//////////////////////////////////////////////////////////////////////////

extern HINSTANCE g_hInst;
extern InputFileOptions *g_pInputOpts;

// prefs.cpp
// Note: The following will generate an error if VIRTUALDUB_MPEG2
// isn't defined for the "VirtualDub" project; this is intentional.
extern int VDPreferencesGetMPEG2Prefs1();
// Main.cpp
extern void CPUTest();

// Preference masks for VirtualDub-MPEG2 (see prefs.cpp)
#define MPEG2_SHOW_SMPTE		1		// show SMPTE time code on timeline
#define MPEG2_PREVIEW_ADJUST	2		// adjust panes to MPEG aspect ratio
#define MPEG2_ALLOW_DSC			4		// allow Direct Stream Copy of MPEG audio
#define MPEG2_DISABLE_RFF		8		// do not honor "repeat first field" flags
#define MPEG2_ALLOW_MATRIX		16		// allow "matrix coefficients" for RGB

//////////////////////////////////////////////////////////////////////////

namespace {
	enum { kVDST_Mpeg = 5 };

	enum {
		kVDM_AudioConcealingError,		// MPEGAudio: Concealing decoding error on frame %lu: %hs.
		kVDM_OpeningFile,				// MPEG: Opening file "%hs"
		kVDM_OOOTimestamp,				// MPEG: Anachronistic or discontinuous timestamp found in %ls stream %d at byte position %lld, from %lld to %lld. This may indicate an improper join.
		kVDM_Incomplete,				// MPEG: File ended unexpectedly during parsing -- file may be damaged or incomplete.
		kVDM_InvalidFramerate,			// MPEG: Video stream has an invalid frame rate code (%i).
		kVDM_TransportStream			// MPEG: File matches the profile of a Transport Stream and may not decode correctly.
	};
}

//////////////////////////////////////////////////////////////////////////

// Remember chosen audio stream (for jobs)?
#define REMEMBER_AUDIO

// Adjust skew transparently?
//#define ADJUST_SKEW

// Store presentation time stamps?
#define STORE_TIME_STAMPS


#define VIDPKT_TYPE_SEQUENCE_START		0x1B3
#define VIDPKT_TYPE_SEQUENCE_END		0x1B7
#define VIDPKT_TYPE_GROUP_START			0x1B8
#define VIDPKT_TYPE_PICTURE_START		0x100
#define VIDPKT_TYPE_SLICE_START_MIN		0x101
#define VIDPKT_TYPE_SLICE_START_MAX		0x1AF
#define VIDPKT_TYPE_EXT_START			0x1B5
#define VIDPKT_TYPE_USER_START			0x1B2

#define SEQUENCE_EXTENSION_ID			1
#define SEQUENCE_DISPLAY_EXTENSION_ID	2
#define PICTURE_CODING_EXTENSION_ID		8

#define MPEG_FRAME_TYPE_I	1
#define MPEG_FRAME_TYPE_P	2
#define MPEG_FRAME_TYPE_B	3

#define MPEG_BUFFER_BIDIRECTIONAL	(0)
#define MPEG_BUFFER_BACKWARD		(1)
#define MPEG_BUFFER_FORWARD			(2)
#define MPEG_BUFFER_AUXILIARY		(3)

// Picture structures
#define TOP_FIELD		1
#define BOTTOM_FIELD	2
#define FRAME_PICTURE	3

//////////////////////////////////////////////////////////////////////////
//
//
//					MPEGPacketInfo/MPEGSampleInfo
//
//
//////////////////////////////////////////////////////////////////////////


typedef struct {
	__int64     file_pos;
	__int64     stream_pos;
#ifdef STORE_TIME_STAMPS
	sint64		PTS;
#endif
} MPEGPacketInfo;

// flags for MPEGSampleInfo.bitflags
#define PICTUREFLAG_RFF		1	// repeat_first_field
#define PICTUREFLAG_TFF		2	// top_field_first
#define PICTUREFLAG_PF		4	// progressive_frame

typedef struct {
	__int64		stream_pos;
#ifdef STORE_TIME_STAMPS
	sint64		PTS;
#endif
	long		size;
	union {
		struct {
			char	frame_type;			// I,P,B
			char	picture_structure;	// FRAME/FIELD
			char	bitflags;			// MPEG-2 rff/tff, etc.
			char	fields;				// # of fields
		};

		struct {
			long	header;
		};
	};
} MPEGSampleInfo;


//////////////////////////////////////////////////////////////////////////
//
//
//					InputFileMPEGOptions
//
//
//////////////////////////////////////////////////////////////////////////

class InputFileMPEGOptions : public InputFileOptions {
public:
	enum {
		DECODE_NO_B = 1,
		DECODE_NO_P = 2,
	};

	struct InputFileMPEGOpts {
		int len;
		int iDecodeMode;
		bool fAcceptPartial;
#ifdef REMEMBER_AUDIO
		int iChosenAudio;
#endif
	} opts;


	InputFileMPEGOptions();
	~InputFileMPEGOptions();

	bool read(const char *buf);
	int write(char *buf, int buflen);

	static INT_PTR APIENTRY SetupDlgProc( HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam);
};

InputFileMPEGOptions::InputFileMPEGOptions() {
	opts.len = sizeof(opts);
	opts.iDecodeMode = 0;
	opts.fAcceptPartial = true;
#ifdef REMEMBER_AUDIO
	opts.iChosenAudio = 0;
#endif
}

InputFileMPEGOptions::~InputFileMPEGOptions() {
}

bool InputFileMPEGOptions::read(const char *buf) {
	const InputFileMPEGOpts *pp = (const InputFileMPEGOpts *)buf;

	if (pp->len != sizeof(InputFileMPEGOpts))
		return false;

	opts = *pp;

	return true;
}

int InputFileMPEGOptions::write(char *buf, int buflen) {
	InputFileMPEGOpts *pp = (InputFileMPEGOpts *)buf;

	if (buflen<sizeof(InputFileMPEGOpts))
		return 0;

	opts.len = sizeof(InputFileMPEGOpts);
	*pp = opts;

	return sizeof(InputFileMPEGOpts);
}

///////

INT_PTR APIENTRY InputFileMPEGOptions::SetupDlgProc( HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam) {
	InputFileMPEGOptions *thisPtr = (InputFileMPEGOptions *)GetWindowLongPtr(hDlg, DWLP_USER);

	switch(message) {
		case WM_INITDIALOG:
			SetWindowLongPtr(hDlg, DWLP_USER, lParam);
			CheckDlgButton(hDlg, IDC_MPEG_ALL_FRAMES, TRUE);
			return TRUE;

		case WM_COMMAND:
			if (LOWORD(wParam) == IDCANCEL) {
				if (IsDlgButtonChecked(hDlg, IDC_MPEG_I_FRAMES_ONLY	))
					thisPtr->opts.iDecodeMode = InputFileMPEGOptions::DECODE_NO_B | InputFileMPEGOptions::DECODE_NO_P;

				if (IsDlgButtonChecked(hDlg, IDC_MPEG_IP_FRAMES_ONLY))
					thisPtr->opts.iDecodeMode = InputFileMPEGOptions::DECODE_NO_B;

				if (IsDlgButtonChecked(hDlg, IDC_MPEG_ALL_FRAMES	))
					thisPtr->opts.iDecodeMode = 0;

				thisPtr->opts.fAcceptPartial = true;

				EndDialog(hDlg, 0);
				return TRUE;
			}
			break;
	}

	return FALSE;
}

//////////////////////////////////////////////////////////////////////////
//
//
//					InputFileMPEG
//
//
//////////////////////////////////////////////////////////////////////////

#define AUDIOTYPE_AC3	0x80
#define AUDIOTYPE_LPCM	0xA0
#define AUDIOTYPE_MPEG	0xC0
#define AUDIOTYPE_MASK	0xE0

class AudioSourceMPEG;		// MPEG Layer I, II, or III
class VideoSourceMPEG;
class AudioSourceAC3;		// Dolby Digital
class AudioSourcePCM;		// Linear PCM
class MPEGFileParser;

class InputFileMPEG : public InputFile {

friend VideoSourceMPEG;
friend AudioSourceMPEG;
friend AudioSourceAC3;
friend AudioSourcePCM;
friend MPEGFileParser;

private:
	char *video_packet_buffer;
	long largest_framesize;
	MPEGPacketInfo *video_packet_list;
	MPEGSampleInfo *video_sample_list;
	MPEGPacketInfo *audio_packet_list;
	MPEGSampleInfo *audio_sample_list;
	int packets, apackets;
	int frames, aframes;
	__int64 firstPTS, afirstPTS;
	int last_packet[2];
	int width, height;
	int aspect_ratio;	// aspect_ratio_information
	VDFraction mFrameRate;
	bool fInterleaved, fIsVCD;
	bool progressive_sequence;

	long *video_field_map;
	long fields;

	long audio_first_header;
	long seq_ext;

	int iDecodeMode;
	int iChosenAudio;
	bool fAcceptPartial;

	bool fAllowDSC;
	bool bForceFILM;
	bool fAllowMatrix;

	FastReadStream *pFastRead;
	VDFile mFile;

	static const char szME[];

	void ReadStream(void *buffer, __int64 pos, long len, bool fAudio);

	bool PostInit();

	void setOptions(InputFileOptions *);
	InputFileOptions *createOptions(const char *buf);
	InputFileOptions *promptForOptions(HWND);
public:
	InputFileMPEG();
	~InputFileMPEG();

	void Init(const wchar_t *szFile);
	static void _InfoDlgThread(void *pvInfo);
	static INT_PTR CALLBACK _InfoDlgProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam);
	void InfoDialog(HWND hwndParent);
};


//////////////////////////////////////////////////////////////////////////
//
//
//						VideoSourceMPEG
//
//
//////////////////////////////////////////////////////////////////////////

class VideoSourceMPEG : public VideoSource {
private:
	InputFileMPEG *parentPtr;

	long frame_forw, frame_back, frame_bidir, frame_aux;

	bool mbFBValid;

	vdrefptr<IVDMPEGDecoder> mpDecoder;

	uint32 mAccelerationFlags;

	void DecodeFrameBuffer(int buffer);

	long prev_IP(long f);
	long next_IP(long f);

	VDPosition prev_field_IP(VDPosition f64);
	VDPosition next_field_IP(VDPosition f64);

	unsigned char *read_buffer1;
	long buffer1_frame;
	unsigned char *read_buffer2;
	long buffer2_frame;

	void UpdateAcceleration();

public:
	VideoSourceMPEG(InputFileMPEG *, IVDMPEGDecoder *mpDecoder);
	~VideoSourceMPEG();

	bool init();
	char getFrameTypeChar(VDPosition lFrameNum);
	bool _isKey(VDPosition lSample);
	virtual VDPosition nearestKey(VDPosition lSample);
	virtual VDPosition prevKey(VDPosition lSample);
	virtual VDPosition nextKey(VDPosition lSample);
	bool setTargetFormat(int depth);
	void invalidateFrameBuffer();
	bool isFrameBufferValid();
	void streamBegin(bool fRealTime, bool bForceReset);
	void streamEnd();
	void streamSetDesiredFrame(VDPosition frame_num);
	VDPosition streamGetNextRequiredFrame(bool& is_preroll);
	int streamGetRequiredCount(long *);
	const void *streamGetFrame(const void *inputBuffer, uint32 data_len, bool is_preroll, VDPosition frame_num);
	const void *getFrame(VDPosition frameNum);
	eDropType getDropType(VDPosition);
	int _read(VDPosition lStart, uint32 lCount, void *lpBuffer, uint32 cbBuffer, uint32 *lBytesRead, uint32 *lSamplesRead);

	VDPosition streamToDisplayOrder(VDPosition sample_num) {
		return sample_num;
	}

	VDPosition displayToStreamOrder(VDPosition display_num) {
		return display_num;
	}

	bool isDecodable(VDPosition sample_num);

	int getMPEGAspectRatio();
};

VideoSourceMPEG::VideoSourceMPEG(InputFileMPEG *parent, IVDMPEGDecoder *pDecoder)
	: mpDecoder(pDecoder)
	, mAccelerationFlags((uint32)-1)
{
	read_buffer1 = NULL;
	read_buffer2 = NULL;
	parentPtr = parent;
}

// Fix for Kylie: limit what res we will support
#define MPEG_MAX_WIDTH		1920
#define MPEG_MAX_HEIGHT		1088
// 1920*1080*4 = 8MB buffer

bool VideoSourceMPEG::init() {
	BITMAPINFOHEADER *bmih;
	int w, h;

	mbFBValid = false;
	frame_forw = frame_back = frame_bidir = frame_aux = -1;

	mSampleFirst = 0;
	mSampleLast = parentPtr->fields;

	if (mSampleLast <= mSampleFirst) return false;

	// Check width and height

	w = (parentPtr->width+15) & -16;
	h = (parentPtr->height+1) & -2;

	if (w <= 0 || h <= 0 ||
		w > MPEG_MAX_WIDTH ||
		h > MPEG_MAX_HEIGHT) return false;

	if (!mpDecoder->Init(w, h, parentPtr->seq_ext))
		throw MyMemoryError();

	mpDecoder->EnableMatrixCoefficients(parentPtr->fAllowMatrix);

	if (!AllocFrameBuffer(w * h * 4 + 4))
		throw MyMemoryError();

	if (!(bmih = (BITMAPINFOHEADER *)allocFormat(sizeof(BITMAPINFOHEADER))))
		throw MyMemoryError();

	if (!(read_buffer1 = new unsigned char[parentPtr->largest_framesize * 2]))
		throw MyMemoryError();

	read_buffer2 = read_buffer1 + parentPtr->largest_framesize;

	buffer1_frame = -1;
	buffer2_frame = -1;

	mpTargetFormatHeader.resize(getFormatLen());

	bmih->biSize			= sizeof(BITMAPINFOHEADER);
	bmih->biWidth			= w;
	bmih->biHeight			= h;
	bmih->biPlanes			= 1;
	bmih->biBitCount		= 32;
	bmih->biCompression		= 0xffffffff;
	bmih->biSizeImage		= 0;
	bmih->biXPelsPerMeter	= 0;
	bmih->biYPelsPerMeter	= 0;
	bmih->biClrUsed			= 0;
	bmih->biClrImportant	= 0;

	streamInfo.fccType					= streamtypeVIDEO;
	streamInfo.fccHandler				= 0;
	streamInfo.dwFlags					= 0;
	streamInfo.wPriority				= 0;
	streamInfo.wLanguage				= 0;
	streamInfo.dwScale					= parentPtr->mFrameRate.getLo();
	streamInfo.dwRate					= parentPtr->mFrameRate.getHi();
	streamInfo.dwStart					= 0;
	streamInfo.dwLength					= (DWORD)mSampleLast;
	streamInfo.dwInitialFrames			= 0;
	streamInfo.dwSuggestedBufferSize	= 0;
	streamInfo.dwQuality				= 0xffffffffL;
	streamInfo.dwSampleSize				= 0;
	streamInfo.rcFrame.left				= 0;
	streamInfo.rcFrame.top				= 0;
	streamInfo.rcFrame.right			= w;
	streamInfo.rcFrame.bottom			= h;

	return true;
}

VideoSourceMPEG::~VideoSourceMPEG() {
	delete [] read_buffer1;
}

bool VideoSourceMPEG::setTargetFormat(int format) {
	using namespace nsVDPixmap;

	if (!format)
		format = kPixFormat_XRGB8888;

	switch(format) {
	case kPixFormat_XRGB1555:
	case kPixFormat_RGB565:
	case kPixFormat_RGB888:
	case kPixFormat_XRGB8888:
	case kPixFormat_YUV422_UYVY:
	case kPixFormat_YUV422_YUYV:
	case kPixFormat_YUV420_Planar:
	case kPixFormat_Y8:
		return VideoSource::setTargetFormat(format);
	}

	return false;
}

void VideoSourceMPEG::invalidateFrameBuffer() {
	mbFBValid = false;
}

bool VideoSourceMPEG::isFrameBufferValid() {
	return mbFBValid;
}

int VideoSourceMPEG::getMPEGAspectRatio() {
	if (VDPreferencesGetMPEG2Prefs1() & MPEG2_PREVIEW_ADJUST) {
		return parentPtr->aspect_ratio;
	}
	return 0;
}

char VideoSourceMPEG::getFrameTypeChar(VDPosition lFrameNum64) {
	if (lFrameNum64 >= mSampleFirst && lFrameNum64 < mSampleLast) {

		long f1 = parentPtr->video_field_map[lFrameNum64];

		if (f1 & 1) {
			return 'R';		// reconstructed
		} else {
			int i = parentPtr->video_sample_list[f1 >> 1].frame_type;
			if (i == MPEG_FRAME_TYPE_I) return 'I';
			if (i == MPEG_FRAME_TYPE_P) return 'P';
			if (i == MPEG_FRAME_TYPE_B) return 'B';
		}
	}
	return ' ';
}

VideoSource::eDropType VideoSourceMPEG::getDropType(VDPosition lFrameNum64) {
	if (lFrameNum64 >= mSampleFirst && lFrameNum64 < mSampleLast) {

		long f1 = parentPtr->video_field_map[lFrameNum64];

		if (f1 & 1) return kDependant;

		switch (parentPtr->video_sample_list[f1 >> 1].frame_type) {
		case MPEG_FRAME_TYPE_I:
			return kIndependent;
		case MPEG_FRAME_TYPE_P:
			return kDependant;
		}
	}
	return kDroppable;
}

bool VideoSourceMPEG::isDecodable(VDPosition sample_num64) {

VDDEBUG("MPEG::isDecodable(%I64i)\n", sample_num64);

	if (sample_num64 >= mSampleFirst && sample_num64 < mSampleLast) {
		// FIX THIS
		/*
		switch (parentPtr->video_sample_list[sample_num64].frame_type) {
		case MPEG_FRAME_TYPE_B:
			if (mpDecoder->GetFrameBuffer(next_IP((long)sample_num64)) < 0) return false;
		case MPEG_FRAME_TYPE_P:
			if (mpDecoder->GetFrameBuffer(prev_IP((long)sample_num64)) < 0) return false;
		}
		*/
		return true;
	}
	return false;
}

bool VideoSourceMPEG::_isKey(VDPosition lSample) {
	if (lSample >= mSampleFirst && lSample < mSampleLast) {
		long f1 = parentPtr->video_field_map[lSample];

		if ((f1 & 1) == 0) {
			if (parentPtr->video_sample_list[f1 >> 1].frame_type == MPEG_FRAME_TYPE_I) {
				return true;
			}
		}
	}
	return false;
}

VDPosition VideoSourceMPEG::nearestKey(VDPosition lSample) {
	if (_isKey(lSample)) return lSample;
	return prevKey(lSample);
}

VDPosition VideoSourceMPEG::prevKey(VDPosition lSample) {
	if (lSample < mSampleLast) {
		while (--lSample >= mSampleFirst) {
			if (_isKey(lSample)) return lSample;
		}
	}
	return -1;
}

VDPosition VideoSourceMPEG::nextKey(VDPosition lSample) {
	if (lSample >= mSampleFirst) {
		while (++lSample < mSampleLast) {
			if (_isKey(lSample)) return lSample;
		}
	}
	return -1;
}

void VideoSourceMPEG::streamBegin(bool fRealTime, bool bForceReset) {

//	_RPT0(0, "VideoSourceMPEG::streamBegin\n");

	UpdateAcceleration();
}

void VideoSourceMPEG::streamEnd() {
	// We are transitioning from predicted frames to
	// current frames, so we must invalidate everything.
	frame_forw = frame_back = frame_bidir = frame_aux = -1;
}


//  streamSetDesiredFrame() Remarks:
//      This function is called once for every frame that
//      the Dub engine wants to get from the decoder.

// Important: We must not modify the buffer tracking
// variables (frame_forw, frame_back, and frame_bidir)!

VDPosition VideoSourceMPEG::prev_field_IP(VDPosition f64) {

	if (f64 < mSampleLast) {
		const MPEGSampleInfo *m1;
		long f1;

		while (--f64 >= mSampleFirst) {
			f1 = parentPtr->video_field_map[f64] >> 1;
			m1 = &(parentPtr->video_sample_list[f1]);
			if (m1->frame_type != MPEG_FRAME_TYPE_B) return f64;
		}
	}

    return -1;
}

VDPosition VideoSourceMPEG::next_field_IP(VDPosition f64) {

	if (f64 >= mSampleFirst) {
		const MPEGSampleInfo *m1;
		long f1;

		while (++f64 < mSampleLast) {
			f1 = parentPtr->video_field_map[f64] >> 1;
			m1 = &(parentPtr->video_sample_list[f1]);
			if (m1->frame_type != MPEG_FRAME_TYPE_B) return f64;
		}
	}

    return -1;
}

// Here is the sequence from "Clip.m2v":
// Frame 10 is P (8)
// Frame 11 is PI (8 + 9)
// Frame 12 is IB (9 + 10)
// Frame 13 is B (10)

//MPEG::streamSetDesiredFrame(10) current = 10, desired = 10
//MPEG::streamGetNextRequiredFrame (current = 10, desired = 10, fwd = 6, back = 8, bidir = -1)
//MPEG::streamSetDesiredFrame: Found cadence PI (8 + 9)
//MPEG::streamSetDesiredFrame(11) current = 11, desired = 11
//MPEG::streamGetNextRequiredFrame (current = 11, desired = 11, fwd = 8, back = 9, bidir = -1)
//MPEG::streamSetDesiredFrame: Found cadence IB (9 + 10)
//MPEG::streamSetDesiredFrame(12) current = 15, desired = 12
//MPEG::streamGetNextRequiredFrame (current = 12, desired = 12, fwd = 9, back = 12, bidir = 10)
//MPEG::streamSetDesiredFrame: Found cadence PI (8 + 9)
//MPEG::streamSetDesiredFrame(11) current = 0, desired = 11
//MPEG::streamGetNextRequiredFrame (current = 11, desired = 11, fwd = 8, back = 9, bidir = 10)
// The killer
//MPEG::streamSetDesiredFrame: Found cadence IB (9 + 10)
//MPEG::streamSetDesiredFrame(12) current = 15, desired = 12
//MPEG::streamGetNextRequiredFrame (current = 15, desired = 12, fwd = 8, back = 9, bidir = 10)
//MPEG::_read (lStart = 15, lCount = 1, lpBuffer = 16744688)
//MPEG::streamGetFrame (frame_num = 15, is_preroll = 1)
//MPEG::streamGetNextRequiredFrame (current = 12, desired = 12, fwd = 9, back = 12, bidir = 10)
//MPEG::streamGetNextRequiredFrame (returning -1)

void VideoSourceMPEG::streamSetDesiredFrame(VDPosition frame_num64) {
	const MPEGSampleInfo *m1;
	long f1, f2, prev, next;
	VDPosition n;

	stream_desired_frame = stream_current_frame = frame_num64;

	f1 = parentPtr->video_field_map[frame_num64];

	// If the desired frame is laced... (aargh)

	if ((f1 & 1) && stream_current_frame > mSampleFirst) {
		const MPEGSampleInfo *m2;

		f2 = f1 >> 1;
		f1 = parentPtr->video_field_map[stream_current_frame - 1] >> 1;

		m1 = &(parentPtr->video_sample_list[f1]);
		m2 = &(parentPtr->video_sample_list[f2]);

// VdubCrash2.mpg:
// Frame 116 is a B frame   (93)
// Frame 117 has cadence BP (93 + 94)
// Frame 118 has cadence PP (94 + 95)
// Frame 119 is a P frame   (95)

#ifdef _DEBUG
		{
			char c1, c2;

			switch (m1->frame_type) {
			case MPEG_FRAME_TYPE_I: c1 = 'I'; break;
			case MPEG_FRAME_TYPE_P: c1 = 'P'; break;
			case MPEG_FRAME_TYPE_B: c1 = 'B'; break;
			}

			switch (m2->frame_type) {
			case MPEG_FRAME_TYPE_I: c2 = 'I'; break;
			case MPEG_FRAME_TYPE_P: c2 = 'P'; break;
			case MPEG_FRAME_TYPE_B: c2 = 'B'; break;
			}

VDDEBUG("MPEG::streamSetDesiredFrame: Found cadence %c%c (%ld + %ld)\n", c1, c2, f1, f2);
		}
#endif

		// There are a lot of possibilities here!

		// Just remember, the question we are answering
		// is whether we need to move stream_current_frame
		// from its original position at stream_desired_frame

		if (m1->frame_type == MPEG_FRAME_TYPE_I) {
			// II or IP, do nothing

			if (m2->frame_type == MPEG_FRAME_TYPE_B) {
				// IB

				// Decode f1, then future IP, then f2

// Fix for Clip.m2v?  We need not be concerned with I, only B:
//				if (frame_forw != f1 || frame_bidir != f2) {
				if (frame_bidir != f2) {

					// Find the future reference
					n = next_field_IP(stream_current_frame);
					if (n >= 0) {
						next = parentPtr->video_field_map[n] >> 1;

						if (frame_forw != f1 || frame_back != next) {
							if (frame_back == f1) {
								// Start at next reference
								stream_current_frame = n;
							} else {
								// Start at I frame
								stream_current_frame--;
							}
						}
					}
				}
			}

		} else if (m1->frame_type == MPEG_FRAME_TYPE_P) {

			if (m2->frame_type != MPEG_FRAME_TYPE_B) {
				// PI, or PP

				// Decode previous reference, then f1, then f2

				// If we don't have f1, back to prev I/P
				// Otherwise start from f2

				if ((frame_forw != f1 || frame_back != f2)
						&& frame_back != f1) {

					// Start from f1 if possible
					stream_current_frame--;
					while (stream_current_frame > mSampleFirst) {
						n = prev_field_IP(stream_current_frame);
						if (n < 0) break;	// ouch
						prev = parentPtr->video_field_map[n] >> 1;
						if (frame_back == prev) break;
						stream_current_frame = n;
						if (parentPtr->video_sample_list[prev].frame_type == MPEG_FRAME_TYPE_I) break;
					}
				}

			} else {
				// PB

				// Decode previous reference, then f1,
				// then future reference, then f2

				if (frame_forw != f1 && frame_back != f1) {

					// Missing the P, go back to prev I/P
					stream_current_frame--;
					while (stream_current_frame > mSampleFirst) {
						n = prev_field_IP(stream_current_frame);
						prev = parentPtr->video_field_map[n] >> 1;
						if (frame_back == prev) break;
						stream_current_frame = n;
						if (parentPtr->video_sample_list[prev].frame_type == MPEG_FRAME_TYPE_I) break;
					}

				} else if (frame_bidir != f2) {
					// Got P, but missing B

					// We'll need the future reference
					n = next_field_IP(stream_current_frame);
					if (n >= 0) {
						next = parentPtr->video_field_map[n] >> 1;

					// If P is in forward buffer but we don't have
					// next, start at P.  (Not sure that can happen)

					// If P is in back buffer and we don't have next
					// then start at next.
						if (frame_forw == f1 && frame_back != next) {
							// Something got screwed
							stream_current_frame--;
							while (stream_current_frame > mSampleFirst) {
								n = prev_field_IP(stream_current_frame);
								prev = parentPtr->video_field_map[n] >> 1;
								if (frame_back == prev) break;
								stream_current_frame = n;
								if (parentPtr->video_sample_list[prev].frame_type == MPEG_FRAME_TYPE_I) break;
							}
						} else if (frame_back == f1) {
							stream_current_frame = n;
						}
					}
				}
			}

		} else {
			// BI, BP, or BB

			if (m2->frame_type != MPEG_FRAME_TYPE_B) {
				// BI, or BP

				// Fix for VdubCrash2.mpg
				if (frame_bidir != f1 || (frame_forw != f2 && frame_back != f2 && m2->frame_type != MPEG_FRAME_TYPE_I)) {

					// Do we have the previous reference?
					n = prev_field_IP(stream_current_frame);
					if (n >= 0) {
						prev = parentPtr->video_field_map[n] >> 1;

						// If we don't have it, start there
						if (frame_back != prev) {

							stream_current_frame = n;
							while (stream_current_frame > mSampleFirst) {
								n = prev_field_IP(stream_current_frame);
								prev = parentPtr->video_field_map[n] >> 1;
								if (frame_back == prev) break;
								stream_current_frame = n;
								if (parentPtr->video_sample_list[prev].frame_type == MPEG_FRAME_TYPE_I) break;
							}

						}
					} else {
						// Degenerate; no previous reference
						stream_current_frame--;
					}
				}

			} else {
				// BB (special case)

				if (frame_bidir != f1 || frame_aux != f2) {
					// We're going to need to look for 2 refs
					VDPosition n2;

					// Find the references
					n = prev_field_IP(stream_current_frame);
					n2 = next_field_IP(stream_current_frame);

					if (n >= 0 && n2 >= 0) {

						prev = parentPtr->video_field_map[n] >> 1;
						next = parentPtr->video_field_map[n2] >> 1;

						if (frame_forw == prev && frame_back == next) {
							// Wow, we have both references!
							// Start from f1 if we don't have it

//							if (frame_bidir != f1) {
//							// (So far, only seen when moving backwards)
//								stream_current_frame--;
//							}

						} else if (frame_back == prev) {
							// We have the prev reference, start at next
							stream_current_frame = n2;

						} else {
							// Drat, go back to prev I/P

							// (This can happen on a random seek)
							stream_current_frame = n;
							while (stream_current_frame > mSampleFirst) {
								n = prev_field_IP(stream_current_frame);
								prev = parentPtr->video_field_map[n] >> 1;
								if (frame_back == prev) break;
								stream_current_frame = n;
								if (parentPtr->video_sample_list[prev].frame_type == MPEG_FRAME_TYPE_I) break;
							}
						}
					}
				}
			}
		}

VDDEBUG("MPEG::streamSetDesiredFrame(%I64i) current = %I64i, desired = %I64i\n", frame_num64, stream_current_frame, stream_desired_frame);
		return;
	}


	// No lacing involved, decode as usual

	f1 >>= 1;
	m1 = &(parentPtr->video_sample_list[f1]);

VDDEBUG("MPEG::streamSetDesiredFrame(%I64i) current = %I64i, desired = %I64i\n", frame_num64, stream_current_frame, stream_desired_frame);

	switch(m1->frame_type) {

	case MPEG_FRAME_TYPE_P:
		if (frame_back != f1 && frame_forw != f1) {
			// Back to previous I/P
			while (stream_current_frame > mSampleFirst) {
				n = prev_field_IP(stream_current_frame);
				if (n < 0) break;	// ouch
				f1 = parentPtr->video_field_map[n] >> 1;
				if (frame_back == f1) break;
				stream_current_frame = n;
				if (parentPtr->video_sample_list[f1].frame_type == MPEG_FRAME_TYPE_I) break;
			}
		}
		break;

	case MPEG_FRAME_TYPE_B:

		if (frame_bidir != f1) {
			VDPosition np, nn;

			// Back to previous I/P

			// Find the past reference
			np = prev_field_IP(stream_current_frame);
			if (np >= 0) prev = parentPtr->video_field_map[np] >> 1;

			// Find the future reference
			nn = next_field_IP(stream_current_frame);
			if (nn >= 0) next = parentPtr->video_field_map[nn] >> 1;

			if (np >= 0 && nn >= 0) {

			// We only need to move if we're missing one of these
			if (frame_forw != prev || frame_back != next) {
				if (frame_back != prev) {
					// Back to prev I/P
					stream_current_frame = np;

					while (stream_current_frame > mSampleFirst) {
						n = prev_field_IP(stream_current_frame);
						if (n < 0) break;	// ouch
						f1 = parentPtr->video_field_map[n] >> 1;
						if (frame_back == f1) break;
						stream_current_frame = n;
						if (parentPtr->video_sample_list[f1].frame_type == MPEG_FRAME_TYPE_I) break;
					}
				} else {
					// Already got prev, start at next
					stream_current_frame = nn;
				}
			}

			}
		}
	}
}

//	streamGetNextRequiredFrame() remarks:
//
//	This function is called repeatedly until it returns -1.
//
//	With each call, the return value is the frame number to read
//	from the file, decode, and possibly display. The reading and
//	decoding stages are skipped (I think) if this function returns
//	-1 the very first time it's called after the dubber specifies
//	its "desired frame."
//
//	The frame numbers returned are queued by the dubber, and
//	eventually passed in sequence to streamGetFrame(). However,
//	it seems both _read() and streamGetFrame() may be called from
//	different thread contexts than streamGetNextRequiredFrame(),
//	thus they cannot share information.
//
//	"is_preroll" should be set to true if the frame number
//	returned is not the "desired frame."  In all other cases
//	it should be returned as false.

/*
This is what happens decoding frame 1 of an IBBP sequence:

MPEG::streamSetDesiredFrame 1
MPEG::streamGetNextRequiredFrame (current = 0, desired = 1, fwd = -1, back = -1, bidir = -1)
MPEG::streamGetNextRequiredFrame (returning 0, is_preroll = 1)
MPEG::_read (lStart = 0, lCount = 1)
MPEG::streamGetFrame (frame_num = 0, is_preroll = 1)
MPEG::streamGetNextRequiredFrame (current = 3, desired = 1, fwd = -1, back = 0, bidir = -1)
MPEG::streamGetNextRequiredFrame (returning 3, is_preroll = 1)
MPEG::_read (lStart = 3, lCount = 1)
MPEG::streamGetFrame (frame_num = 3, is_preroll = 1)
MPEG::streamGetNextRequiredFrame (current = 1, desired = 1, fwd = 0, back = 3, bidir = -1)
MPEG::streamGetNextRequiredFrame (returning 1, is_preroll = 0)
MPEG::_read (lStart = 1, lCount = 1)
MPEG::streamGetFrame (frame_num = 1, is_preroll = 0)
MPEG::streamGetNextRequiredFrame (current = 1, desired = 1, fwd = 0, back = 3, bidir = 1)
MPEG::streamGetNextRequiredFrame (returning -1, is_preroll = 0)

// Then, stepping back to frame 0
MPEG::streamSetDesiredFrame 0
MPEG::streamGetNextRequiredFrame (current = 0, desired = 0, fwd = 0, back = 3, bidir = 1)
MPEG::streamGetNextRequiredFrame (returning -1, is_preroll = 0)
MPEG::streamGetFrame (frame_num = 0, is_preroll = 0)
*/

VDPosition VideoSourceMPEG::streamGetNextRequiredFrame(bool& is_preroll) {
	const MPEGSampleInfo *m1, *m2;
	long f1, f2;
	VDPosition current;

VDDEBUG("MPEG::streamGetNextRequiredFrame (current = %I64i,"
	" desired = %I64i, fwd = %d, back = %d, bidir = %d)\n",
	stream_current_frame, stream_desired_frame,
	frame_forw, frame_back, frame_bidir);

	current = stream_current_frame;

	f1 = parentPtr->video_field_map[current];

	// Only decode I/P frames, until we reach the desired frame

	if (current < stream_desired_frame) {
		f1 >>= 1;
		if (parentPtr->video_sample_list[f1].frame_type != MPEG_FRAME_TYPE_B) {
			if (frame_back != f1) {
				frame_forw = frame_back;
				frame_back = f1;
			}
		}
		stream_current_frame = next_field_IP(current);

		// Watch out for this weird situation:
		if (stream_current_frame == (stream_desired_frame - 1)) {
			f1 = parentPtr->video_field_map[stream_current_frame];
			f2 = parentPtr->video_field_map[stream_desired_frame];
			if ((f1 & 1) == 1 && f1 == (f2 + 1)) {
				// We don't need f1, so skip to desired
				stream_current_frame = stream_desired_frame;
			}
		}

		is_preroll = true;
		return current;
	}

	if (current > stream_desired_frame) {
		// We're one frame away from being finished
		f1 >>= 1;
		if (parentPtr->video_sample_list[f1].frame_type != MPEG_FRAME_TYPE_B) {
			if (frame_back != f1) {
				frame_forw = frame_back;
				frame_back = f1;
			}
		}
		stream_current_frame = stream_desired_frame;
		is_preroll = true;
		return current;
	}

	// We've reached the desired frame.  But in the
	// case of a laced frame, we must emulate everything
	// streamGetFrame does to ensure it can be assembled.

	if ((f1 & 1) && current > mSampleFirst) {
		f2 = f1 >> 1;
		f1 = parentPtr->video_field_map[current - 1] >> 1;

		m1 = &(parentPtr->video_sample_list[f1]);
		m2 = &(parentPtr->video_sample_list[f2]);

		if (m1->frame_type != MPEG_FRAME_TYPE_B) {

			if (m2->frame_type != MPEG_FRAME_TYPE_B) {
				// II, IP, PI, PP

				// Do we have f1?
				if (frame_forw == f1) {
					if (frame_back == f2) goto Neg1;
					frame_back = f2;
				} else if (frame_back == f1) {
					// Hmmm, got f1 in wrong buffer
					frame_back = frame_forw;
					frame_forw = f1;
					if (frame_back == f2) goto Neg1;
					frame_back = f2;
				} else if (frame_forw == f2) {
					// Not got f1, but got f2 in wrong buffer
					frame_forw = f1;
					frame_back = f2;
				} else if (frame_back == f2) {
					// Huh? Not got f1, but f2 in correct buffer!
					// (streamGetFrame will take care of this)
					frame_forw = f1;
				} else {
					// Ain't got neither f1 nor f2 (triple negative)
					frame_forw = f1;
					frame_back = f2;
				}

			} else {
				// IB or PB
				if (frame_forw != f1) {
					if (frame_back == f1) {
						// Hmm, f1 in wrong buffer
						frame_back = frame_forw;
						frame_forw = f1;
					} else {
						frame_back = f1;
					}
				}
				if (frame_bidir == f2) goto Neg1;
				frame_bidir = f2;
			}
		} else {
			if (m2->frame_type != MPEG_FRAME_TYPE_B) {
				// BI or BP
				if (frame_back != f2) {
					frame_forw = frame_back;
					frame_back = f2;
				}
				if (frame_bidir == f1) goto Neg1;
				frame_bidir = f1;
			} else {
				// BB (special case)
				if (frame_bidir == f1 && frame_aux == f2) goto Neg1;
				frame_bidir = f1;
				frame_aux = f2;
			}
		}


	} else {
		// No lacing, thank goodness
		f1 >>= 1;
		m1 = &(parentPtr->video_sample_list[f1]);

		if (m1->frame_type != MPEG_FRAME_TYPE_B) {
			if (frame_back == f1) goto Neg1;
//			if (frame_forw == f1) goto Neg1;
			frame_forw = frame_back;
			frame_back = f1;
		} else {
			if (frame_bidir == f1) goto Neg1;
			frame_bidir = f1;
		}
	}

	is_preroll = false;
VDDEBUG("MPEG::streamGetNextRequiredFrame (returning %I64i, preroll = %d)\n", current, is_preroll);
	return current;

Neg1:
	// Finished!

	frame_aux = -1;
	is_preroll = false;
VDDEBUG("MPEG::streamGetNextRequiredFrame (returning -1)\n");
	return -1;
}

int VideoSourceMPEG::streamGetRequiredCount(long *pSize) {

// Return int = the number of frames which need to be decoded
// Return *pSize = the amount of memory needed to decode them

// Perhaps the easiest way to do this is to emulate
// VirtualDub calls to streamGetNextRequiredFrame.

// First, save ALL tracking variables...

	VDPosition saved_stream_current_frame = stream_current_frame;
	VDPosition saved_stream_desired_frame = stream_desired_frame;
	long saved_frame_forw = frame_forw;
	long saved_frame_back = frame_back;
	long saved_frame_bidir = frame_bidir;
	long saved_frame_aux = frame_aux;

	long size = 0;
	int needed = 0;

	for (;;) {
		bool p;
		VDPosition next = streamGetNextRequiredFrame(p);
		if (next == -1) break;

		long f1 = parentPtr->video_field_map[next];
		size += parentPtr->video_sample_list[f1 >> 1].size;

		if ((f1 & 1) && next > mSampleFirst) {
			f1 = parentPtr->video_field_map[next - 1] >> 1;
			size += parentPtr->video_sample_list[f1].size;
		}
		needed++;
	}

	stream_current_frame = saved_stream_current_frame;
	stream_desired_frame = saved_stream_desired_frame;
	frame_forw = saved_frame_forw;
	frame_back = saved_frame_back;
	frame_bidir = saved_frame_bidir;
	frame_aux = saved_frame_aux;

	if (pSize) *pSize = size;
	return needed;
}

const void *VideoSourceMPEG::streamGetFrame(const void *inputBuffer, uint32 data_len, bool is_preroll, VDPosition frame_num64) {
	const MPEGSampleInfo *m1, *m2;
	long f1, f2;
	int buffer, b1, b2;

VDDEBUG("MPEG::streamGetFrame (frame_num = %I64i, is_preroll = %d)\n", frame_num64, is_preroll);

	if (frame_num64 < mSampleFirst) return NULL;

	f1 = parentPtr->video_field_map[frame_num64];

	if (is_preroll) {
		// Just decode, but skip B-frames

		// If this is a laced frame, we actually want the second!
		if ((f1 & 1) && frame_num64 > mSampleFirst) {
			f2 = f1 >> 1;
			f1 = parentPtr->video_field_map[frame_num64 - 1] >> 1;

			m1 = &(parentPtr->video_sample_list[f1]);
			m2 = &(parentPtr->video_sample_list[f2]);

VDASSERT(data_len >= (m1->size + m2->size));

			buffer = mpDecoder->GetFrameBuffer(f2);

			if (m2->frame_type != MPEG_FRAME_TYPE_B) {
				if (buffer != MPEG_BUFFER_BACKWARD) {
					mpDecoder->SwapFrameBuffers(MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
					if (buffer != MPEG_BUFFER_FORWARD) {
						mpDecoder->DecodeFrame((char *)inputBuffer + m1->size, m2->size, f2, MPEG_BUFFER_BACKWARD, MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
					}
				}
			}

		} else {
			f1 >>= 1;
			m1 = &(parentPtr->video_sample_list[f1]);
			buffer = mpDecoder->GetFrameBuffer(f1);

VDASSERT(data_len >= m1->size);

			if (m1->frame_type != MPEG_FRAME_TYPE_B) {
				if (buffer != MPEG_BUFFER_BACKWARD) {
					mpDecoder->SwapFrameBuffers(MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
					if (buffer != MPEG_BUFFER_FORWARD) {
						mpDecoder->DecodeFrame(inputBuffer, data_len, f1, MPEG_BUFFER_BACKWARD, MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
					}
				}
			}
		}
	}

	if (!is_preroll) {
		// Decode, assemble, and display

		if ((f1 & 1) && frame_num64 > mSampleFirst) {

			f2 = f1 >> 1;
			f1 = parentPtr->video_field_map[frame_num64 - 1] >> 1;

			m1 = &(parentPtr->video_sample_list[f1]);
			m2 = &(parentPtr->video_sample_list[f2]);

			b1 = mpDecoder->GetFrameBuffer(f1);
			b2 = mpDecoder->GetFrameBuffer(f2);

			if (m1->frame_type != MPEG_FRAME_TYPE_B) {

				if (m2->frame_type != MPEG_FRAME_TYPE_B) {
					// II, IP, PI, PP

					// Do we have f1?
					if (b1 == MPEG_BUFFER_FORWARD) {
						// Maybe we also have f2?
						if (b2 != MPEG_BUFFER_BACKWARD) {
VDASSERT(data_len >= (m1->size + m2->size));
							mpDecoder->DecodeFrame((char *)inputBuffer + m1->size, m2->size, f2, MPEG_BUFFER_BACKWARD, MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
						}
					} else if (b1 == MPEG_BUFFER_BACKWARD) {
						// Hmmm, got f1 in other buffer
						mpDecoder->SwapFrameBuffers(MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
						if (b2 != MPEG_BUFFER_FORWARD) {
VDASSERT(data_len >= (m1->size + m2->size));
							mpDecoder->DecodeFrame((char *)inputBuffer + m1->size, m2->size, f2, MPEG_BUFFER_BACKWARD, MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
						}
					} else if (b2 == MPEG_BUFFER_FORWARD) {
						// Not got f1, but got f2 in other buffer
VDASSERT(data_len >= (m1->size + m2->size));
						mpDecoder->DecodeFrame(inputBuffer, m1->size, f1, MPEG_BUFFER_BACKWARD, MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
						mpDecoder->SwapFrameBuffers(MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
					} else if (b2 == MPEG_BUFFER_BACKWARD) {
						// Not got f1, but got f2
VDASSERT(data_len >= (m1->size + m2->size));
						mpDecoder->DecodeFrame(inputBuffer, m1->size, f1, MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD, MPEG_BUFFER_FORWARD);
					} else {
						// Ain't got neither
VDASSERT(data_len >= (m1->size + m2->size));
						mpDecoder->DecodeFrame(inputBuffer, m1->size, f1, MPEG_BUFFER_BACKWARD, MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
						mpDecoder->SwapFrameBuffers(MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
						mpDecoder->DecodeFrame((char *)inputBuffer + m1->size, m2->size, f2, MPEG_BUFFER_BACKWARD, MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
					}

				} else {
					// IB, PB
					if (b1 != MPEG_BUFFER_FORWARD) {
						if (b1 == MPEG_BUFFER_BACKWARD) {
							// Hmm, f1 in wrong buffer
							mpDecoder->SwapFrameBuffers(MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
						} else {
VDASSERT(data_len >= (m1->size + m2->size));
							mpDecoder->DecodeFrame(inputBuffer, m1->size, f1, MPEG_BUFFER_BACKWARD, MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
						}
					}
					if (b2 != MPEG_BUFFER_BIDIRECTIONAL) {
VDASSERT(data_len >= (m1->size + m2->size));
						mpDecoder->DecodeFrame((char *)inputBuffer + m1->size, m2->size, f2, MPEG_BUFFER_BIDIRECTIONAL, MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
					}
				}
			} else {
				if (m2->frame_type != MPEG_FRAME_TYPE_B) {
					// BI, BP (degenerate)
					if (b2 != MPEG_BUFFER_BACKWARD) {
						mpDecoder->SwapFrameBuffers(MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
						if (b2 != MPEG_BUFFER_FORWARD) {
VDASSERT(data_len >= (m1->size + m2->size));
							mpDecoder->DecodeFrame((char *)inputBuffer + m1->size, m2->size, f2, MPEG_BUFFER_BACKWARD, MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
						}
					}
					if (b1 != MPEG_BUFFER_BIDIRECTIONAL) {
VDASSERT(data_len >= (m1->size + m2->size));
						mpDecoder->DecodeFrame(inputBuffer, m1->size, f1, MPEG_BUFFER_BIDIRECTIONAL, MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
					}
				} else {
					// BB
					if (b1 != MPEG_BUFFER_BIDIRECTIONAL) {
VDASSERT(data_len >= (m1->size + m2->size));
						mpDecoder->DecodeFrame(inputBuffer, m1->size, f1, MPEG_BUFFER_BIDIRECTIONAL, MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
					}
					if (b2 != MPEG_BUFFER_AUXILIARY) {
VDASSERT(data_len >= (m1->size + m2->size));
						mpDecoder->DecodeFrame((char *)inputBuffer + m1->size, m2->size, f2, MPEG_BUFFER_AUXILIARY, MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
					}
				}
			}

			// Now assemble the fields into auxiliary buffer
			b1 = mpDecoder->GetFrameBuffer(f1);
			b2 = mpDecoder->GetFrameBuffer(f2);

VDASSERT(b1 >= 0);

			if (m2->frame_type == MPEG_FRAME_TYPE_B) {
				if (m1->frame_type == MPEG_FRAME_TYPE_B) {
					buffer = MPEG_BUFFER_AUXILIARY;
				} else {
					buffer = MPEG_BUFFER_BIDIRECTIONAL;
				}
			} else {
				buffer = MPEG_BUFFER_BACKWARD;
			}

VDASSERT(b2 == buffer);

			if (m1->picture_structure != FRAME_PICTURE) {
				if (m1->picture_structure == TOP_FIELD) {
					// m1 top + m2 bottom
					mpDecoder->CopyField(MPEG_BUFFER_AUXILIARY, 1, buffer, 1);
					mpDecoder->CopyField(MPEG_BUFFER_AUXILIARY, 0, b1, 0);
				} else {
					// m1 bottom + m2 top
					mpDecoder->CopyField(MPEG_BUFFER_AUXILIARY, 0, buffer, 0);
					mpDecoder->CopyField(MPEG_BUFFER_AUXILIARY, 1, b1, 1);
				}

			} else if (m2->picture_structure != FRAME_PICTURE) {
				if (m2->picture_structure == TOP_FIELD) {
					// m1 bottom + m2 top
					mpDecoder->CopyField(MPEG_BUFFER_AUXILIARY, 0, buffer, 0);
					mpDecoder->CopyField(MPEG_BUFFER_AUXILIARY, 1, b1, 1);
				} else {
					// m1 top + m2 bottom
					mpDecoder->CopyField(MPEG_BUFFER_AUXILIARY, 1, buffer, 1);
					mpDecoder->CopyField(MPEG_BUFFER_AUXILIARY, 0, b1, 0);
				}

			} else {
				if (m1->bitflags & PICTUREFLAG_RFF) {
					if (m1->bitflags & PICTUREFLAG_TFF) {
						// m1 top + m2 bottom
						mpDecoder->CopyField(MPEG_BUFFER_AUXILIARY, 1, buffer, 1);
						mpDecoder->CopyField(MPEG_BUFFER_AUXILIARY, 0, b1, 0);
					} else {
						// m1 bottom + m2 top
						mpDecoder->CopyField(MPEG_BUFFER_AUXILIARY, 0, buffer, 0);
						mpDecoder->CopyField(MPEG_BUFFER_AUXILIARY, 1, b1, 1);
					}
				} else {
					if (m1->bitflags & PICTUREFLAG_TFF) {
						// m1 bottom + m2 top
						mpDecoder->CopyField(MPEG_BUFFER_AUXILIARY, 0, buffer, 0);
						mpDecoder->CopyField(MPEG_BUFFER_AUXILIARY, 1, b1, 1);
					} else {
						// m1 top + m2 bottom
						mpDecoder->CopyField(MPEG_BUFFER_AUXILIARY, 1, buffer, 1);
						mpDecoder->CopyField(MPEG_BUFFER_AUXILIARY, 0, b1, 0);
					}
				}
			}

			buffer = MPEG_BUFFER_AUXILIARY;

		} else {
			// Frame picture
			f1 >>= 1;
			m1 = &(parentPtr->video_sample_list[f1]);
			b1 = mpDecoder->GetFrameBuffer(f1);

			if (m1->frame_type != MPEG_FRAME_TYPE_B) {
				if (b1 == MPEG_BUFFER_FORWARD) {
					// Already got f1 in fwd buffer
					buffer = MPEG_BUFFER_FORWARD;
				} else {
					// Got f1 in back buffer?
					if (b1 != MPEG_BUFFER_BACKWARD) {
						mpDecoder->SwapFrameBuffers(MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
						if (b1 != MPEG_BUFFER_FORWARD) {
							mpDecoder->DecodeFrame(inputBuffer, m1->size, f1, MPEG_BUFFER_BACKWARD, MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
						}
					}
					buffer = MPEG_BUFFER_BACKWARD;
				}
			} else {
				if (b1 != MPEG_BUFFER_BIDIRECTIONAL) {
VDASSERT(data_len >= m1->size);
					mpDecoder->DecodeFrame(inputBuffer, m1->size, f1, MPEG_BUFFER_BIDIRECTIONAL, MPEG_BUFFER_FORWARD, MPEG_BUFFER_BACKWARD);
				}
				buffer = MPEG_BUFFER_BIDIRECTIONAL;
			}
		}

		DecodeFrameBuffer(buffer);
		mbFBValid = true;
	}

#ifndef _M_AMD64
//	if (MMX_enabled) __asm emms
#endif

	return lpvBuffer;
}

const void *VideoSourceMPEG::getFrame(VDPosition frameNum64) {
	int buffer;

VDDEBUG("MPEG::getFrame (frameNum = %I64i)\n", frameNum64);

	UpdateAcceleration();

	if (frameNum64 < mSampleFirst || frameNum64 >= mSampleLast)
		throw MyError("VideoSourceMPEG->getFrame(): Invalid frameNum");

	// Do we have the buffer stored somewhere?
	buffer = mpDecoder->GetFrameBuffer((long)frameNum64);

	if (buffer < 0) {	// nope
		VDPosition next;
		uint32 br, sr;
		bool p;

		//streamEnd();
		streamSetDesiredFrame(frameNum64);
		next = streamGetNextRequiredFrame(p);
		if (next == -1) {
			this->read(frameNum64, 1, parentPtr->video_packet_buffer, parentPtr->largest_framesize * 2, &br, &sr);
			streamGetFrame(parentPtr->video_packet_buffer, br, p, frameNum64);
		} else {
			do {
				this->read(next, 1, parentPtr->video_packet_buffer, parentPtr->largest_framesize * 2, &br, &sr);
				streamGetFrame(parentPtr->video_packet_buffer, br, p, next);
				next = streamGetNextRequiredFrame(p);
			} while (next != -1);
		}

	} else {
		DecodeFrameBuffer(buffer);
	}

#ifndef _M_AMD64
	//if (MMX_enabled) __asm emms
#endif

	mbFBValid = true;
	return getFrameBuffer();
}

void VideoSourceMPEG::DecodeFrameBuffer(int buffer) {
	char *pBuffer = (char *)lpvBuffer;
	const long w = getImageFormat()->biWidth;
	const long h = getImageFormat()->biHeight;

	VDASSERT(buffer >= 0);

	using namespace nsVDPixmap;

	switch(mTargetFormat.format) {
	case kPixFormat_YUV420_Planar:
		{
			ptrdiff_t ypitch, cbpitch, crpitch;
			const void *py = mpDecoder->GetYBuffer(buffer, ypitch);
			const void *pcr = mpDecoder->GetCrBuffer(buffer, crpitch);
			const void *pcb = mpDecoder->GetCbBuffer(buffer, cbpitch);
			VDMemcpyRect(pBuffer, w, py, ypitch, w, h);
			pBuffer += w*h;
			VDMemcpyRect(pBuffer, w>>1, pcr, crpitch, w >> 1, h >> 1);
			pBuffer += w*h >> 2;
			VDMemcpyRect(pBuffer, w>>1, pcb, cbpitch, w >> 1, h >> 1);
		}
		break;

	case kPixFormat_Y8:
		{
			ptrdiff_t ypitch;
			const void *py = mpDecoder->GetYBuffer(buffer, ypitch);
			VDMemcpyRect(pBuffer, w, py, ypitch, w, h);
		}
		break;

	case kPixFormat_YUV422_UYVY:
		mpDecoder->DecodeUYVY(pBuffer, w*2, buffer);
		break;
	case kPixFormat_YUV422_YUYV:
		mpDecoder->DecodeYUYV(pBuffer, w*2, buffer);
		break;
	case kPixFormat_XRGB1555:
		mpDecoder->DecodeRGB15(pBuffer + w*2*(h-1), -w*2, buffer);
		break;
	case kPixFormat_RGB565:
		mpDecoder->DecodeRGB16(pBuffer + w*2*(h-1), -w*2, buffer);
		break;
	case kPixFormat_RGB888:
		mpDecoder->DecodeRGB24(pBuffer + w*3*(h-1), -w*3, buffer);
		break;
	case kPixFormat_XRGB8888:
		mpDecoder->DecodeRGB32(pBuffer + w*4*(h-1), -w*4, buffer);
		break;
	}
}

int VideoSourceMPEG::_read(VDPosition lStart, uint32 lCount, void *lpBuffer, uint32 cbBuffer, uint32 *lBytesRead, uint32 *lSamplesRead) {
	const MPEGSampleInfo *m1, *m2;
	long f1, f2, len;
	int cadence;

	if (lStart < mSampleFirst || lStart >= mSampleLast)
		return AVIERR_BADPARAM;

	f1 = parentPtr->video_field_map[lStart];
	cadence = f1 & 1;
	f1 >>= 1;

	if (cadence && lStart > mSampleFirst) {
		f2 = f1;
		f1 = parentPtr->video_field_map[lStart - 1] >> 1;

		m1 = &(parentPtr->video_sample_list[f1]);
		m2 = &(parentPtr->video_sample_list[f2]);

		len = m1->size + m2->size;
	} else {

		m1 = &(parentPtr->video_sample_list[f1]);
		len = m1->size;
	}

	if (lCount == AVISTREAMREAD_CONVENIENT) lCount = 1;

VDDEBUG("MPEG::_read (lStart = %I64i, lCount = %d, lpBuffer = %d)\n", lStart, lCount, (long)lpBuffer);

	if (!lpBuffer) {
		if (lSamplesRead) *lSamplesRead = 1;
		if (lBytesRead) *lBytesRead = len;
		return AVIERR_OK;
	}

	if (len > cbBuffer) {
		if (lSamplesRead) *lSamplesRead = 0;
		if (lBytesRead) *lBytesRead = 0;
		return AVIERR_BUFFERTOOSMALL;
	}

	if (cadence) {
		if (buffer1_frame == f1) {
			memcpy(lpBuffer, read_buffer1, m1->size);
			if (buffer2_frame != f2) {
				parentPtr->ReadStream(read_buffer2, m2->stream_pos, m2->size, false);
				buffer2_frame = f2;
			}
			memcpy((char *)lpBuffer + m1->size, read_buffer2, m2->size);

		} else if (buffer2_frame == f1) {
			memcpy(lpBuffer, read_buffer2, m1->size);
			if (buffer1_frame != f2) {
				parentPtr->ReadStream(read_buffer1, m2->stream_pos, m2->size, false);
				buffer1_frame = f2;
			}
			memcpy((char *)lpBuffer + m1->size, read_buffer1, m2->size);

		} else if (buffer1_frame == f2) {
			parentPtr->ReadStream(read_buffer2, m1->stream_pos, m1->size, false);
			memcpy(lpBuffer, read_buffer2, m1->size);
			buffer2_frame = f1;
			memcpy((char *)lpBuffer + m1->size, read_buffer1, m2->size);

		} else if (buffer2_frame == f2) {
			parentPtr->ReadStream(read_buffer1, m1->stream_pos, m1->size, false);
			memcpy(lpBuffer, read_buffer1, m1->size);
			buffer1_frame = f1;
			memcpy((char *)lpBuffer + m1->size, read_buffer2, m2->size);

		} else {
			parentPtr->ReadStream(read_buffer1, m1->stream_pos, m1->size, false);
			parentPtr->ReadStream(read_buffer2, m2->stream_pos, m2->size, false);
			buffer1_frame = f1;
			buffer2_frame = f2;
			memcpy(lpBuffer, read_buffer1, m1->size);
			memcpy((char *)lpBuffer + m1->size, read_buffer2, m2->size);
		}

	} else {
		if (buffer1_frame == f1) {
			memcpy(lpBuffer, read_buffer1, m1->size);
		} else if (buffer2_frame == f1) {
			memcpy(lpBuffer, read_buffer2, m1->size);
		} else {
			parentPtr->ReadStream(read_buffer2, m1->stream_pos, m1->size, false);
			memcpy(lpBuffer, read_buffer2, m1->size);
			buffer2_frame = f1;
		}
	}

	if (lSamplesRead) *lSamplesRead = 1;
	if (lBytesRead) *lBytesRead = len;
	return AVIERR_OK;
}

///////

long VideoSourceMPEG::prev_IP(long lSample) {
	if (lSample < mSampleLast) {
		while (--lSample >= mSampleFirst) {
			if (parentPtr->video_sample_list[lSample].frame_type
				!= MPEG_FRAME_TYPE_B) return lSample;
		}
	}
    return -1;
}

long VideoSourceMPEG::next_IP(long lSample) {
	if (lSample >= mSampleFirst) {
		while (++lSample < mSampleLast) {
			if (parentPtr->video_sample_list[lSample].frame_type
			  != MPEG_FRAME_TYPE_B) return lSample;
		}
	}
    return -1;
}

void VideoSourceMPEG::UpdateAcceleration() {
	uint32 flags = CPUGetEnabledExtensions();

	if (mAccelerationFlags != flags) {
		mAccelerationFlags = flags;

		static const uint32 sse2_flags = (CPUF_SUPPORTS_SSE2 | CPUF_SUPPORTS_INTEGER_SSE | CPUF_SUPPORTS_MMX);
		static const uint32 isse_flags = (CPUF_SUPPORTS_INTEGER_SSE | CPUF_SUPPORTS_MMX);
		static const uint32 mmx_flags = CPUF_SUPPORTS_MMX;

#ifdef _M_AMD64
//		mpDecoder->SetPredictors(&g_VDMPEGPredict_sse2);
//		mpDecoder->SetConverters(&g_VDMPEGConvert_reference);
		mpDecoder->SetIDCTs(&g_VDMPEGIDCT_sse2);
#else
		if ((flags & sse2_flags) == sse2_flags) {
//			mpDecoder->SetPredictors(&g_VDMPEGPredict_sse2);
//			mpDecoder->SetConverters(&g_VDMPEGConvert_isse);
			mpDecoder->SetIDCTs(&g_VDMPEGIDCT_isse);
		} else if ((flags & isse_flags) == isse_flags) {
//			mpDecoder->SetPredictors(&g_VDMPEGPredict_isse);
//			mpDecoder->SetConverters(&g_VDMPEGConvert_isse);
			mpDecoder->SetIDCTs(&g_VDMPEGIDCT_isse);
		} else if ((flags & mmx_flags) == mmx_flags) {
//			mpDecoder->SetPredictors(&g_VDMPEGPredict_mmx);
//			mpDecoder->SetConverters(&g_VDMPEGConvert_mmx);
			mpDecoder->SetIDCTs(&g_VDMPEGIDCT_mmx);
		} else {
//			mpDecoder->SetPredictors(&g_VDMPEGPredict_scalar);
//			mpDecoder->SetConverters(&g_VDMPEGConvert_scalar);
			mpDecoder->SetIDCTs(&g_VDMPEGIDCT_scalar);
		}
#endif
	}
}

//////////////////////////////////////////////////////////////////////////
//
//
//						AudioSourceMPEG
//
//
//////////////////////////////////////////////////////////////////////////

static const int bitrate[2][3][16] = {
		{
			{ 0, 32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448, 0 },	// MPEG-1 layer I
			{ 0, 32, 48, 56,  64,  80,  96, 112, 128, 160, 192, 224, 256, 320, 384, 0 },	// MPEG-1 layer II
			{ 0, 32, 40, 48,  56,  64,  80,  96, 112, 128, 160, 192, 224, 256, 320, 0 },	// MPEG-1 layer III
		},
		{
			{ 0, 32, 48, 56,  64,  80,  96, 112, 128, 144, 160, 176, 192, 224, 256, 0 },	// MPEG-2 layer I
			{ 0,  8, 16, 24,  32,  40,  48,  56,  64,  80,  96, 112, 128, 144, 160, 0 },	// MPEG-2 layer II
			{ 0,  8, 16, 24,  32,  40,  48,  56,  64,  80,  96, 112, 128, 144, 160, 0 },	// MPEG-2 layer III
		}
};

static const long samp_freq[2][2][4] = {
	{
		{ 44100, 48000, 32000, 0 },		// MPEG-1
		{ 22050, 24000, 16000, 0 },		// MPEG-2
	},
	{
		{     0,     0,     0, 0 },		// impossible
		{ 11025, 12000,  8000, 0 },		// MPEG-2.5
	}
};

// MPEG audio frame header (bswapped):

// ........ ........ 111..... 11111111 syncword
// ........ ........ ...xx... ........ version (11=1, 10=2, 00=2.5)
// ........ ........ .....xx. ........ layer (01=III, 10=II, 11=I)
// ........ ........ .......x ........ crc (0=yes, 1=no)
// xx...... ........ ........ ........ mode (00=stereo, 10=joint, 01=dual, 11=mono)
// ..xx.... ........ ........ ........ mode ext (only for joint stereo)
// ....x... ........ ........ ........ copyright (0=no, 1=yes)
// .....x.. ........ ........ ........ original (0=orig, 1=copy)
// ......xx ........ ........ ........ emphasis (not 10)
// ........ xxxx.... ........ ........ bitrate (not 0000 nor 1111)
// ........ ....xx.. ........ ........ sampling rate (not 11)
// ........ ......x. ........ ........ padded (0=no, 1=yes)
// ........ .......x ........ ........ private bit

struct MPEGAudioHeader {
	enum {
		kMaskNone		= 0x00000000,
		kMaskSync		= 0x0000E0FF,
		kMaskMPEG25		= 0x00001000,
		kMaskVersion	= 0x00000800,
		kMaskLayer		= 0x00000600,
		kMaskCRC		= 0x00000100,
		kMaskBitrate	= 0x00F00000,
		kMaskSampleRate	= 0x000C0000,
		kMaskPadding	= 0x00020000,
		kMaskPrivate	= 0x00010000,
		kMaskMode		= 0xC0000000,
		kMaskModeExt	= 0x30000000,
		kMaskCopyright	= 0x08000000,
		kMaskOriginal	= 0x04000000,
		kMaskEmphasis	= 0x03000000,
		kMaskAll		= 0xFFFFFFFF
	};

	const uint32 mHeader;

	MPEGAudioHeader(uint32 hdr) : mHeader(hdr) {}

	bool		IsSyncValid() const				{ return (mHeader & kMaskSync) == kMaskSync; }
	bool		IsMPEG2() const					{ return !(mHeader & kMaskVersion); }
	bool		IsMPEG25() const				{ return IsMPEG2() && !(mHeader & kMaskMPEG25); }
	unsigned	GetLayer() const				{ return 4 - ((mHeader >> 9)&3); }
	bool		IsCRCProtected() const			{ return !(mHeader & kMaskCRC); }
	unsigned	GetBitrateIndex() const			{ return (mHeader >> 20) & 15; }
	unsigned	GetSamplingRateIndex() const	{ return (mHeader >> 18) & 3; }
	bool		IsPadded() const				{ return 0!=(mHeader & kMaskPadding); }
//	unsigned    GetModeIndex() const            { return (mHeader >> 28) & 3; }
	bool		IsStereo() const				{ return (mHeader & kMaskMode) != kMaskMode; }

	unsigned	GetBitrateKbps() const			{ return bitrate[IsMPEG2()][GetLayer()-1][GetBitrateIndex()]; }
	unsigned	GetSamplingRateHz() const		{ return samp_freq[IsMPEG25()][IsMPEG2()][GetSamplingRateIndex()]; }

// Added these:
	bool		IsCopyrighted() const			{ return ((mHeader & kMaskCopyright)!=0); }
	bool		IsOriginal() const				{ return ((mHeader & kMaskOriginal)==0); }
	bool		IsPrivate() const				{ return ((mHeader & kMaskPrivate)!=0); }
	unsigned	GetModeExt() const				{ return (mHeader >> 28) & 3; }
	unsigned	GetMode() const					{ return (mHeader >> 30) & 3; }
	unsigned	GetEmphasis() const				{ return (mHeader >> 24) & 3; }

	bool IsValid() const;
	bool IsConsistent(uint32 hdr) const;
	unsigned GetFrameSize() const;
	unsigned GetPayloadSizeL3() const;
};

bool MPEGAudioHeader::IsValid() const {
	return IsSyncValid()					// need at least 11 bits for sync mark
		&& (!IsMPEG25() || IsMPEG2())		// either twelfth bit is set or it's MPEG-2.5
		&& GetLayer() != 4					// layer IV invalid
		&& GetBitrateIndex() != 15			// bitrate=1111 invalid
		&& GetSamplingRateIndex() != 3		// sampling_rate=11 reserved
		;
}

bool MPEGAudioHeader::IsConsistent(uint32 hdr) const {
	uint32 headerdiff = (hdr ^ mHeader);

	// do not allow MPEG version, layer, or sampling rate to change
	if (headerdiff & (kMaskSampleRate | kMaskVersion | kMaskLayer | kMaskMPEG25))
		return false;

	// only layer III decoders must accept VBR
	if (GetLayer() != 3 && (headerdiff & kMaskBitrate))
		return false;

	return true;
}

unsigned MPEGAudioHeader::GetFrameSize() const {
	const bool		is_mpeg2	= IsMPEG2();
	const unsigned	bitrate		= GetBitrateKbps();
	const unsigned	freq		= GetSamplingRateHz();
	const unsigned	padding		= IsPadded();

	if (GetLayer() == 1)
		return 4*(12000*bitrate/freq + padding);
	else {
		if (is_mpeg2)
			return (72000*bitrate/freq + padding);
		else
			return (144000*bitrate/freq + padding);
	}
}

unsigned MPEGAudioHeader::GetPayloadSizeL3() const {
	VDASSERT(GetLayer() == 3);

	static const unsigned sideinfo_size[2][2]={17,32,9,17};

	unsigned size = GetFrameSize() - sideinfo_size[IsMPEG2()][IsStereo()];

	if (IsCRCProtected())
		size -= 2;

	VDASSERT((signed)size > 0);

	return size;
}

class AudioSourceMPEG : public AudioSource, IAMPBitsource {
private:
	InputFileMPEG *parentPtr;
	IAMPDecoder *iad;
	void *pkt_buffer;
	short sample_buffer[1152*2][2];
	char *pDecoderPoint;
	char *pDecoderLimit;

	long lCurrentPacket;
	int layer;
	int mSamplesPerFrame;

	bool fAllowDSC;
	bool mbIsMPEG2;
	ErrorMode mErrorMode;

	BOOL _isKey(LONG lSample);

public:
	AudioSourceMPEG(InputFileMPEG *);
	~AudioSourceMPEG();

	// AudioSource methods

	bool init();
	int _read(VDPosition lStart, uint32 lCount, void *lpBuffer, uint32 cbBuffer, uint32 *lBytesRead, uint32 *lSamplesRead);

	ErrorMode getDecodeErrorMode() { return mErrorMode; }
	void setDecodeErrorMode(ErrorMode mode);
	bool isDecodeErrorModeSupported(ErrorMode mode);

	// IAMPBitsource methods

	int read(void *buffer, int bytes);

};

AudioSourceMPEG::AudioSourceMPEG(InputFileMPEG *pp)
	: AudioSource()
	, mErrorMode(kErrorModeReportAll)
{
	parentPtr = pp;

	if (!(pkt_buffer = new char[8192]))
		throw MyMemoryError();

	if (!(iad = CreateAMPDecoder())) {
		delete [] pkt_buffer;
		throw MyMemoryError();
	}

	this->fAllowDSC = false;
	lCurrentPacket = -1;
}

AudioSourceMPEG::~AudioSourceMPEG() {
	delete [] pkt_buffer;
	iad->Destroy();
}

BOOL AudioSourceMPEG::_isKey(LONG lSample) {
	return TRUE;
}

bool AudioSourceMPEG::init() {
	WAVEFORMATEX *wfex;

	MPEGAudioHeader	header(parentPtr->audio_first_header);

	layer = header.GetLayer();

	mSamplesPerFrame = (layer==1 ? 384 : layer == 3 && header.IsMPEG2() ? 576 : 1152);

	mSampleFirst = 0;

	this->fAllowDSC = (parentPtr->fAllowDSC &&
		(layer == 3 || !header.IsMPEG2()));

	// Allow Direct Stream Copy of audio stream?
	if (this->fAllowDSC) {    

		int brate = header.GetBitrateKbps();
		long srate = header.GetSamplingRateHz();
		int version = header.IsMPEG2()? 2: 1;

		mSampleLast = (parentPtr->aframes > 0)?
			parentPtr->audio_sample_list[parentPtr->aframes-1].stream_pos
			+ parentPtr->audio_sample_list[parentPtr->aframes-1].size: 0;

		if (layer == 3) {
			MPEGLAYER3WAVEFORMAT *wfex3;

			if (!(wfex3 = (MPEGLAYER3WAVEFORMAT *)allocFormat(sizeof(MPEGLAYER3WAVEFORMAT))))
				return false;

			wfex3->wfx.wFormatTag		= WAVE_FORMAT_MPEGLAYER3;
			wfex3->wfx.nChannels		= (WORD)(header.IsStereo()? 2: 1);
			wfex3->wfx.nSamplesPerSec	= srate;
			wfex3->wfx.nBlockAlign		= 1;
			wfex3->wfx.wBitsPerSample	= 0;
			wfex3->wfx.cbSize			= MPEGLAYER3_WFX_EXTRA_BYTES;
			wfex3->wID					= MPEGLAYER3_ID_MPEG;
			wfex3->fdwFlags				= MPEGLAYER3_FLAG_PADDING_ISO;  //?
			wfex3->nBlockSize			= (WORD)(((144000 / version) * brate) / srate);
			wfex3->nFramesPerBlock		= 1;
			wfex3->nCodecDelay			= 1393;
			wfex3->wfx.nAvgBytesPerSec	= (srate * version * wfex3->nBlockSize) / 1152;

			streamInfo.fccType				= streamtypeAUDIO;
			streamInfo.fccHandler			= 0;
			streamInfo.dwCaps				= 0;
			streamInfo.wPriority			= 0;
			streamInfo.wLanguage			= 0;
			streamInfo.dwFlags				= 0;
			streamInfo.dwScale				= 1;
			streamInfo.dwRate				= wfex3->wfx.nAvgBytesPerSec;
			streamInfo.dwStart				= 0;
			streamInfo.dwLength				= (long)mSampleLast;
			streamInfo.dwInitialFrames		= 1;
			streamInfo.dwSuggestedBufferSize= 0;
			streamInfo.dwQuality			= 0xffffffffL;
			streamInfo.dwSampleSize			= 1;
			streamInfo.rcFrame.left			= 0;
			streamInfo.rcFrame.top			= 0;
			streamInfo.rcFrame.right		= 0;
			streamInfo.rcFrame.bottom		= 0;

		} else {    // Layer I or II, MPEG-1 only
			MPEG1WAVEFORMAT *wfex1;
			double total_br;
			int i;

			bool pad = header.IsPadded();

			int layermask[4] = {
				ACM_MPEG_LAYER1,
				ACM_MPEG_LAYER2,
				ACM_MPEG_LAYER3,
				0};

			int headmode[4] = {
				ACM_MPEG_STEREO,
				ACM_MPEG_JOINTSTEREO,
				ACM_MPEG_DUALCHANNEL,
				ACM_MPEG_SINGLECHANNEL};

			if (!(wfex1 = (MPEG1WAVEFORMAT *)allocFormat(sizeof(MPEG1WAVEFORMAT))))
				return false;

			wfex1->wfx.wFormatTag		= WAVE_FORMAT_MPEG;
			wfex1->wfx.nChannels		= (WORD)(header.IsStereo()? 2: 1);
			wfex1->wfx.nSamplesPerSec	= srate;
			wfex1->wfx.nBlockAlign		= 1;
			wfex1->wfx.wBitsPerSample	= 0;
			wfex1->wfx.cbSize			= 22;
			wfex1->dwHeadBitrate		= brate * 1000;

			total_br = 0.0;
			wfex1->fwHeadMode = 0;
			wfex1->fwHeadFlags = 0;
			wfex1->fwHeadLayer = 0;

			// Assume a fixed frame size
			if (layer == 1) {
				wfex1->wfx.nBlockAlign = (WORD)
					(4 * (int)(12000 * brate / srate));
			} else {	// Layer II
				wfex1->wfx.nBlockAlign = (WORD)
					(144000 * brate / srate);
			}

			for (i=0; i<parentPtr->aframes; i++) {
				MPEGAudioHeader	hdr(parentPtr->audio_sample_list[i].header);

				// Channel mode may change
				int k = headmode[hdr.GetMode()];
				wfex1->fwHeadMode |= k;
				if (k == ACM_MPEG_JOINTSTEREO) {
					wfex1->fwHeadModeExt |= (1 << hdr.GetModeExt());
				}

				// Layer may change
				wfex1->fwHeadLayer |= layermask[hdr.GetLayer()-1];

				// Sampling rate may change
				if (hdr.GetSamplingRateHz() != srate)
					wfex1->wfx.nSamplesPerSec = 0;

				// Bitrate may change
				k = hdr.GetBitrateKbps();
				total_br += k;
				if (k != brate) {
					// Variable Bit Rate
					wfex1->dwHeadBitrate = 0;
					wfex1->wfx.nBlockAlign = 1;
				} else if (hdr.IsPadded() != pad) {
					// Variable frame size
					wfex1->wfx.nBlockAlign = 1;
				}

				// Flags may change
				if (!hdr.IsMPEG2()) wfex1->fwHeadFlags |= ACM_MPEG_ID_MPEG1;
				if (hdr.IsPrivate()) wfex1->fwHeadFlags |= ACM_MPEG_PRIVATEBIT;
				if (hdr.IsCopyrighted()) wfex1->fwHeadFlags |= ACM_MPEG_COPYRIGHT;
				if (hdr.IsOriginal()) wfex1->fwHeadFlags |= ACM_MPEG_ORIGINALHOME;
				if (hdr.IsCRCProtected()) wfex1->fwHeadFlags |= ACM_MPEG_PROTECTIONBIT;
			}

			wfex1->wfx.nAvgBytesPerSec = (parentPtr->aframes<=0)? 0:
				VDRoundToLong((total_br * 125.0) / (double)parentPtr->aframes);

			wfex1->wHeadEmphasis = (WORD)(header.GetEmphasis()+1);

			// Should we store the real PTS?
			wfex1->dwPTSLow = 0;
			wfex1->dwPTSHigh = 0;

			streamInfo.fccType				= streamtypeAUDIO;
			streamInfo.fccHandler			= 0;
			streamInfo.dwCaps				= 0;
			streamInfo.wPriority			= 0;
			streamInfo.wLanguage			= 0;
			streamInfo.dwFlags				= 0;
			streamInfo.dwScale				= wfex1->wfx.nBlockAlign;
			streamInfo.dwRate				= wfex1->wfx.nAvgBytesPerSec;
			streamInfo.dwStart				= 0;
			streamInfo.dwLength				= (DWORD)(mSampleLast/wfex1->wfx.nBlockAlign);
			streamInfo.dwInitialFrames		= 0;
			streamInfo.dwSuggestedBufferSize= 0;
			streamInfo.dwQuality			= 0xffffffffL;
			streamInfo.dwSampleSize			= wfex1->wfx.nBlockAlign;
			streamInfo.rcFrame.left			= 0;
			streamInfo.rcFrame.top			= 0;
			streamInfo.rcFrame.right		= 0;
			streamInfo.rcFrame.bottom		= 0;
		}
		return true;
	}

	mSampleLast = parentPtr->aframes * mSamplesPerFrame;

	if (!(wfex = (WAVEFORMATEX *)allocFormat(sizeof(WAVEFORMATEX))))
		return false;

	// [?] * [bits/sec] / [samples/sec] = [bytes/frame]
	// [?] * [bits/sample] = [bytes/frame]
	// [?] * [.125 byte/bit] * [bits/sample] = [bytes/frame]
	// [samples/frame]

	wfex->wFormatTag		= WAVE_FORMAT_PCM;
	wfex->nChannels			= (WORD)(header.IsStereo()? 2: 1);
	wfex->nSamplesPerSec	= header.GetSamplingRateHz();
	wfex->nBlockAlign		= (WORD)(wfex->nChannels*2);
	wfex->nAvgBytesPerSec	= wfex->nSamplesPerSec * wfex->nBlockAlign;
	wfex->wBitsPerSample	= 16;
	wfex->cbSize			= 0;

	streamInfo.fccType					= streamtypeAUDIO;
	streamInfo.fccHandler				= 0;
	streamInfo.dwCaps					= 0;
	streamInfo.wPriority				= 0;
	streamInfo.wLanguage				= 0;
	streamInfo.dwFlags					= 0;
	streamInfo.dwScale					= wfex->nBlockAlign;
	streamInfo.dwRate					= wfex->nSamplesPerSec * wfex->nBlockAlign;
	streamInfo.dwStart					= 0;
	streamInfo.dwLength					= parentPtr->aframes * mSamplesPerFrame;
	streamInfo.dwInitialFrames			= 0;
	streamInfo.dwSuggestedBufferSize	= 0;
	streamInfo.dwQuality				= 0xffffffffL;
	streamInfo.dwSampleSize				= wfex->nBlockAlign;
	streamInfo.rcFrame.left				= 0;
	streamInfo.rcFrame.top				= 0;
	streamInfo.rcFrame.right			= 0;
	streamInfo.rcFrame.bottom			= 0;

	try {
		iad->Init();
		iad->setSource(this);
	} catch(int i) {
		throw MyError(iad->getErrorString(i));
	}

	return true;
}

int AudioSourceMPEG::_read(VDPosition lStart, uint32 lCount, void *lpBuffer, uint32 cbBuffer, uint32 *lBytesRead, uint32 *lSamplesRead) {
	long lAudioPacket;
	MPEGSampleInfo *msi;
	long samples, len;

	long ba = getWaveFormat()->nBlockAlign;

	if (this->fAllowDSC) {
		if (lCount == AVISTREAMREAD_CONVENIENT)	{
			lCount = (ba > 1)? 1:
				MPEGAudioHeader(parentPtr->audio_first_header).GetFrameSize();
		}
		if (!lpBuffer || lCount <= 0) {
			if (lCount < 0) lCount = 0;
			if (lSamplesRead) *lSamplesRead = lCount;
			if (lBytesRead) *lBytesRead = lCount;
			return AVIERR_OK;
		}
		len = lCount*ba;
		if (len > cbBuffer) {
			if (lSamplesRead) *lSamplesRead = 0;
			if (lBytesRead) *lBytesRead = 0;
			return AVIERR_BUFFERTOOSMALL;
		}
		parentPtr->ReadStream(lpBuffer, lStart*ba, len, true);
		if (lSamplesRead) *lSamplesRead = lCount;
		if (lBytesRead) *lBytesRead = len;
		return AVIERR_OK;
	}

	lAudioPacket = (long)lStart/mSamplesPerFrame;
	samples = mSamplesPerFrame - ((long)lStart % mSamplesPerFrame);

	if (lCount != AVISTREAMREAD_CONVENIENT)
		if (samples > lCount) samples = lCount;

	if (samples*ba > cbBuffer && lpBuffer) {
		samples = cbBuffer / ba;

		if (samples <= 0) {
			if (lSamplesRead) *lSamplesRead = 0;
			if (lBytesRead) *lBytesRead = 0;
			return AVIERR_BUFFERTOOSMALL;
		}
	}

	if (!lpBuffer) {
		if (lSamplesRead) *lSamplesRead = samples;
		if (lBytesRead) *lBytesRead = samples * ba;
		return AVIERR_OK;
	}

	// Because of the overlap from the subband synthesis window, we must
	// decode the previous packet in order to avoid glitches in the output.
	// For layer III, we must also ensure that at least 511 bytes are present
	// in the bit reservoir, as packets may dig that far back using
	// main_data_begin.  An MPEG-1 stereo stream may use up to 32 bytes of
	// the data payload for sideband information.

	if (lCurrentPacket != lAudioPacket) {
		try {
			long nDecodeStart;

			if (layer!=3 || lCurrentPacket<0 || lCurrentPacket+1 != lAudioPacket) {
//				_RPT0(0,"Resetting...\n");
				iad->Reset();

				lCurrentPacket = lAudioPacket;
				--lCurrentPacket;
				if (lCurrentPacket < 0)
					lCurrentPacket = 0;

				nDecodeStart = lCurrentPacket;

				// layer III: add packets to preload bit reservoir
				if (layer == 3) {
					long nReservoirDelay = 511;		// main_data_start is 9 bits (0..511)
					
					while(lCurrentPacket > 0 && nReservoirDelay > 0) {
						--lCurrentPacket;

						nReservoirDelay -= MPEGAudioHeader(parentPtr->audio_sample_list[lCurrentPacket].header).GetPayloadSizeL3();
					}
				}
			} else {
				lCurrentPacket	= lAudioPacket;
				nDecodeStart	= lAudioPacket;
			}

			do {
//				_RPT1(0,"Decoding packet: %d\n", lCurrentPacket);

				msi = &parentPtr->audio_sample_list[lCurrentPacket];
				len = msi->size;

				parentPtr->ReadStream(pkt_buffer, msi->stream_pos, len, TRUE);

				pDecoderPoint = (char *)pkt_buffer;
				pDecoderLimit = (char *)pkt_buffer + len;

				if ((unsigned char)pDecoderPoint[0] != 0xff || ((unsigned char)pDecoderPoint[1]&0xe0)!=0xe0)
					throw MyInternalError("MPEG audio header in sample list has bad sync mark.\n(%s:%d)", __FILE__, __LINE__);

				iad->setDestination((short *)sample_buffer);
				iad->ReadHeader();

				try {
					if (lCurrentPacket < nDecodeStart)
						iad->PrereadFrame();
					else
						iad->DecodeFrame();
				} catch(int i) {
					if (mErrorMode != kErrorModeReportAll) {
						const char *pszError = iad->getErrorString(i);
//						unsigned frame = lAudioPacket;
						VDLogAppMessage(kVDLogWarning, kVDST_Mpeg, kVDM_AudioConcealingError, 2, &lAudioPacket, &pszError);

						if (lCurrentPacket >= nDecodeStart)
							memset(sample_buffer, 0, sizeof sample_buffer);

						iad->ConcealFrame();
					} else
						throw;
				}

			} while(++lCurrentPacket <= lAudioPacket);
			--lCurrentPacket;
		} catch(int i) {
			char buf[64];
			long badpacket = lCurrentPacket;

			uint32 ticks = (uint32)(0.5 + (badpacket * 1000.0 * mSamplesPerFrame) / getWaveFormat()->nSamplesPerSec);
			ticks_to_str(buf, ticks);

			lCurrentPacket = -1;

			throw MyError("Error decoding MPEG audio frame %lu (%s): %s", (unsigned long)badpacket, buf, iad->getErrorString(i));
		}
	}

	memcpy(lpBuffer, (short *)sample_buffer + (lStart%mSamplesPerFrame)*(ba/2), samples*ba);

	if (lSamplesRead) *lSamplesRead = samples;
	if (lBytesRead) *lBytesRead = samples*ba;

	return AVIERR_OK;
}

int AudioSourceMPEG::read(void *buffer, int bytes) {
	if (pDecoderPoint+bytes > pDecoderLimit)
		throw MyError("Incomplete audio frame");

	memcpy(buffer, pDecoderPoint, bytes);
	pDecoderPoint += bytes;

	return bytes;
}

void AudioSourceMPEG::setDecodeErrorMode(ErrorMode mode) {
	if (mode == kErrorModeDecodeAnyway)
		mode = kErrorModeConceal;
	mErrorMode = mode;
}

bool AudioSourceMPEG::isDecodeErrorModeSupported(ErrorMode mode) {
	return mode == kErrorModeReportAll || mode == kErrorModeConceal;
}


//////////////////////////////////////////////////////////////////////////
//
//
//						AudioSourceAC3
//
//
//////////////////////////////////////////////////////////////////////////

// Derived from liba52
static int ac3_syncinfo(void *src, long *srate, int *brate, int *chans) {

	static const int rate[19] = {32,40,48,56,64,80,96,112,
		128,160,192,224,256,320,384,448,512,576,640};
	static const char lfeon[8] = {16,16,4,4,4,1,4,1};
	static const char chan[8] = {2,1,2,3,3,4,4,5};

	const unsigned char *buf = (unsigned char *)src;

	if (*(short *)src == 0x770B && buf[4] < 0xC0) {
		int frmsizecod = buf[4] & 63;
		if (frmsizecod < 38) {
			int bsid = buf[5] >> 3;
			if (bsid < 12) {
				int acmod = buf[6] >> 5;
				int bitrate = rate[frmsizecod >> 1];

				if (bsid > 8) bsid -= 8; else bsid = 0;
				if (chans) *chans = chan[acmod] + !!(buf[6] & lfeon[acmod]);
				if (brate) *brate = bitrate >> bsid;

				switch (buf[4] & 0xC0) {
					case 0:
						if (srate) *srate = 48000 >> bsid;
						return (4 * bitrate);
					case 0x40:
						if (srate) *srate = 44100 >> bsid;
						return (2 * (320 * bitrate / 147 + (frmsizecod & 1)));
					case 0x80:
						if (srate) *srate = 32000 >> bsid;
						return (6 * bitrate);
				}
			}
		}
	}
	return 0;
}

class AudioSourceAC3 : public AudioSource {
private:
	InputFileMPEG *parentPtr;

public:
	AudioSourceAC3(InputFileMPEG *);
	~AudioSourceAC3();

	// AudioSource methods
	bool init();
	int _read(VDPosition lStart, uint32 lCount, void *lpBuffer, uint32 cbBuffer, uint32 *lBytesRead, uint32 *lSamplesRead);
};

AudioSourceAC3::AudioSourceAC3(InputFileMPEG *pp) : AudioSource() {
	parentPtr = pp;
}

AudioSourceAC3::~AudioSourceAC3() {
}

// For AudioSourceAC3, audio_first_header is of this form:
//	........ ........ ........ ..xxxxxx	fscode
//	........ ........ ........ xx......	srcode
//	........ ........ .....xxx ........	audio service
//	........ ........ xxxxx... ........	version
//	........ xxx..... ........ ........	audio coding mode

bool AudioSourceAC3::init() {
	WAVEFORMATEX *wfex;

	unsigned long syncinfo[2];
	int frmsize, chans, brate;
	long srate;

	mSampleFirst = mSampleLast = 0;

	wfex = (WAVEFORMATEX *)allocFormat(sizeof(WAVEFORMATEX)*2);
	if (!wfex) return false;

	syncinfo[0] = 0x770B;
	syncinfo[1] = parentPtr->audio_first_header;

	frmsize = ac3_syncinfo(syncinfo, &srate, &brate, &chans);
	if (!frmsize) return false;

	// Construct WAVEFORMATEX for AC-3
	wfex->wFormatTag = 0x2000;
	wfex->nChannels = (WORD)chans;
	wfex->nSamplesPerSec = srate;
	wfex->nAvgBytesPerSec = brate * 125UL;
	wfex->wBitsPerSample = 0;

	if (srate == 44100) {
		if (parentPtr->aframes > 0) {
			mSampleLast = parentPtr->audio_sample_list[parentPtr->aframes-1].stream_pos
				+ parentPtr->audio_sample_list[parentPtr->aframes-1].size;
		}
		if (mSampleLast < 0) mSampleLast = 0;
		wfex->nBlockAlign = 1;
	} else {
		if (parentPtr->aframes > 0) mSampleLast = parentPtr->aframes;
		wfex->nBlockAlign = (WORD)frmsize;
	}

	// I think someone goofed this
	wfex->cbSize = sizeof(WAVEFORMATEX);
	memset((char *)wfex + sizeof(WAVEFORMATEX), 0, sizeof(WAVEFORMATEX));

	streamInfo.fccType					= streamtypeAUDIO;
	streamInfo.fccHandler				= 0;
	streamInfo.dwCaps					= 0;
	streamInfo.wPriority				= 0;
	streamInfo.wLanguage				= 0;
	streamInfo.dwFlags					= 0;
	streamInfo.dwScale					= wfex->nBlockAlign;
	streamInfo.dwRate					= wfex->nAvgBytesPerSec;
	streamInfo.dwStart					= 0;
	streamInfo.dwLength					= (long)mSampleLast;
	streamInfo.dwInitialFrames			= 0;
	streamInfo.dwSuggestedBufferSize	= 0;
	streamInfo.dwQuality				= 0xffffffffL;
	streamInfo.dwSampleSize				= wfex->nBlockAlign;
	streamInfo.rcFrame.left				= 0;
	streamInfo.rcFrame.top				= 0;
	streamInfo.rcFrame.right			= 0;
	streamInfo.rcFrame.bottom			= 0;

	return true;
}

int AudioSourceAC3::_read(VDPosition lStart, uint32 lCount, void *lpBuffer, uint32 cbBuffer, uint32 *lBytesRead, uint32 *lSamplesRead) {

	const int ba = this->getWaveFormat()->nBlockAlign;

	if (lCount == AVISTREAMREAD_CONVENIENT)
		lCount = (ba == 1)? 2048: 1;

	if (!lpBuffer || lCount <= 0) {
		if (lCount < 0) lCount = 0;
		if (lSamplesRead) *lSamplesRead = lCount;
		if (lBytesRead) *lBytesRead = lCount * ba;
		return AVIERR_OK;
	}

#ifdef ADJUST_SKEW
	// Adjust for skew?
	double skew = ((double)parentPtr->afirstPTS - (double)parentPtr->firstPTS) / 90000.0;
	lStart -= (long)((skew * this->getWaveFormat()->nSamplesPerSec) / 1536.0);
#endif

	if (lStart < mSampleFirst || lStart >= mSampleLast) {
		if (lSamplesRead) *lSamplesRead = 0;
		if (lBytesRead) *lBytesRead = 0;
		return AVIERR_OK;
	}

	if (lSamplesRead) *lSamplesRead = lCount;

	if (ba != 1) {
		lStart = parentPtr->audio_sample_list[lStart].stream_pos;
		lCount *= ba;
	}

	parentPtr->ReadStream(lpBuffer, lStart, lCount, true);
	if (lBytesRead) *lBytesRead = lCount;

	return AVIERR_OK;
}

//////////////////////////////////////////////////////////////////////////
//
//
//						AudioSourcePCM
//
//
//////////////////////////////////////////////////////////////////////////

#define PCM_BUFFER_SIZE	1536	// samples

class AudioSourcePCM : public AudioSource {
private:
	InputFileMPEG *parentPtr;
	void *pkt_buffer;
//	short sample_buffer[PCM_BUFFER_SIZE*2][2];

	long lCurrentPacket;

	inline void LPCM16_to_WAVE16(void *src, int len);
	inline void LPCM20M_to_WAVE16M(void *src, int len);
	inline void LPCM20S_to_WAVE16S(void *src, int len);
	inline void LPCM24M_to_WAVE16M(void *src, int len);
	inline void LPCM24S_to_WAVE16S(void *src, int len);

public:
	AudioSourcePCM(InputFileMPEG *);
	~AudioSourcePCM();

	// AudioSource methods
	bool init();
	int _read(VDPosition lStart, uint32 lCount, void *lpBuffer, uint32 cbBuffer, uint32 *lBytesRead, uint32 *lSamplesRead);
};

AudioSourcePCM::AudioSourcePCM(InputFileMPEG *pp) : AudioSource() {
	parentPtr = pp;

	// The output is in blocks of 1536 samples,
	// always mono or stereo, and always 16-bit...
	// But our input may be multichannel 24-bit audio,
	// so we will need an asymmetric read buffer.

	// The size we need is
	//		PCM_BUFFER_SIZE (1536 samples)
	//		* (maximum number of channels we support)
	//		* (maximum bits per sample / 8)

	// Note that if we drop PCM_BUFFER_SIZE to 192,
	// we can use the same buffer AudioSourceMPEG
	// uses, and still support all possible LPCM:
	//		1152 * 2 * 2 = 4608
	//		192 * 8 * 3 = 4608
	// If we do the conversion in place then
	// we actually don't need the pkt_buffer.

	if (!(pkt_buffer = new char[PCM_BUFFER_SIZE * 2 * 3]))
		throw MyMemoryError();

	lCurrentPacket = -1;
}

AudioSourcePCM::~AudioSourcePCM() {
	delete [] pkt_buffer;
}

bool AudioSourcePCM::init() {

	WAVEFORMATEX *wfex;
	int hdr;

	if (!(wfex = (WAVEFORMATEX*)allocFormat(sizeof(WAVEFORMATEX))))
		return false;

	hdr = parentPtr->audio_first_header;

	// audio_first_header is:
	// 11111111 ........ ........ ........	non-zero
	// ........ x....... ........ ........	emphasis
	// ........ .x...... ........ ........	mute
	// ........ ..x..... ........ ........	reserved
	// ........ ...xxxxx ........ ........	frame number % 20
	// ........ ........ xxxxxxxx ........	dynamic range control
	// ........ ........ ........ xx......	00=16-,01=20-,10=24-bit (11=forbidden)
	// ........ ........ ........ ..xx....	00=48,01=96 KHz (others reserved?)
	// ........ ........ ........ ....x...	reserved
	// ........ ........ ........ .....xxx	number of channels - 1

	wfex->wFormatTag = WAVE_FORMAT_PCM;
	wfex->nChannels = (WORD)((hdr & 7)? 2: 1); //(hdr & 7) + 1;
	wfex->nSamplesPerSec = (hdr & 0x30)? 96000: 48000;
	wfex->wBitsPerSample = 16; //+ ((hdr & 0xC0) >> 4);
	wfex->nBlockAlign = (WORD)(wfex->nChannels * (wfex->wBitsPerSample / 8));
	wfex->nAvgBytesPerSec = wfex->nSamplesPerSec * wfex->nBlockAlign;

	mSampleFirst = 0;
	mSampleLast = parentPtr->aframes * PCM_BUFFER_SIZE;

	streamInfo.fccType					= streamtypeAUDIO;
	streamInfo.fccHandler				= 0;
	streamInfo.dwCaps					= 0;
	streamInfo.wPriority				= 0;
	streamInfo.wLanguage				= 0;
	streamInfo.dwFlags					= 0;
	streamInfo.dwScale					= wfex->nBlockAlign;
	streamInfo.dwRate					= wfex->nAvgBytesPerSec;
	streamInfo.dwStart					= 0;
	streamInfo.dwLength					= (long)mSampleLast;
	streamInfo.dwInitialFrames			= 0;
	streamInfo.dwSuggestedBufferSize	= 0;
	streamInfo.dwQuality				= 0xffffffffL;
	streamInfo.dwSampleSize				= wfex->nBlockAlign;
	streamInfo.rcFrame.left				= 0;
	streamInfo.rcFrame.top				= 0;
	streamInfo.rcFrame.right			= 0;
	streamInfo.rcFrame.bottom			= 0;

	return true;
}

// All of these are "in place" conversions

// None of them are optimized, but clarity
// here was more important than speed <g>

inline void AudioSourcePCM::LPCM16_to_WAVE16(void *src, int len) {

// This is just a straightforward byte swap...

	__asm {
		mov		esi, src
		mov		ecx, len

		sar		ecx, 2
		js		done		; len < 0 not allowed
		jz		part2		; less than 4 bytes left

		align 16
loop1:	mov		eax, [esi]	; [3][2][1][0]
		bswap	eax
		rol		eax, 16
		mov		[esi], eax
		add		esi, 4
		dec		ecx
		jg		loop1

part2:	test	len, 2
		jz		done		; less than 2 bytes left

		mov		ax, word ptr [esi]
		xchg	ah, al
		mov		word ptr [esi], ax
done:
	}
}

inline void AudioSourcePCM::LPCM20S_to_WAVE16S(void *src, int len) {
    __asm {
		mov		esi, src
		mov		ecx, len
		mov		edi, esi

	// This is how the LPCM stream is coded:
	//	buf[0]				=    top 8 bits of Left sample 1
	//  buf[1]				= middle 8 bits of Left sample 1
	//	buf[2]				=    top 8 bits of Right sample 1
	//	buf[3]				= middle 8 bits of Right sample 1
	//	buf[4]				=    top 8 bits of Left sample 2
	//  buf[5]				= middle 8 bits of Left sample 2
	//	buf[6]				=    top 8 bits of Right sample 2
	//	buf[7]				= middle 8 bits of Right sample 2
	//	buf[8] bits [7..4]	= bottom 4 bits of Left sample 1
	//  buf[8] bits [3..0]	= bottom 4 bits of Right sample 1
	//	buf[9] bits [7..4]	= bottom 4 bits of Left sample 2
	//  buf[9] bits [3..0]	= bottom 4 bits of Right sample 2

	// We convert to Intel 16-bit as shown:
	//	buf[0] = buf[1] = middle 8 bits of Left sample 1
	//	buf[1] = buf[0] =    top 8 bits of Left sample 1
	//	buf[2] = buf[3]	= middle 8 bits of Right sample 1
	//	buf[3] = buf[2] =    top 8 bits of Right sample 1
	//	buf[4] = buf[5] = middle 8 bits of Left sample 1
	//	buf[5] = buf[4] =    top 8 bits of Left sample 1
	//	buf[6] = buf[7]	= middle 8 bits of Right sample 1
	//	buf[7] = buf[6] =    top 8 bits of Right sample 1

		align 16
loop1:	sub		ecx, 5
		jl		done

		mov		eax, [esi]		; 3 2 1 0
		add		esi, 4
		bswap	eax				; 0 1 2 3
		ror		eax, 16			; 2 3 0 1
		mov		[edi], eax
		add		edi, 4

		sub		ecx, 5
		jl		done

		mov		eax, [esi]		; 7 6 5 4
		add		esi, 6
		bswap	eax				; 4 5 6 7
		ror		eax, 16			; 6 7 4 5
		mov		[edi], eax
		add		edi, 4

		jmp		loop1
done:
	}
}

inline void AudioSourcePCM::LPCM20M_to_WAVE16M(void *src, int len) {
    __asm {
		mov		esi, src
		mov		ecx, len
		mov		edi, esi

	// This is how the LPCM stream is coded:
	//	buf[0]				=    top 8 bits of Sample 1
	//  buf[1]				= middle 8 bits of Sample 1
	//	buf[2]				=    top 8 bits of Sample 2
	//	buf[3]				= middle 8 bits of Sample 2
	//	buf[4] bits [7..4]	= bottom 4 bits of Sample 1
	//  buf[4] bits [3..0]	= bottom 4 bits of Sample 2

	// We convert to Intel 16-bit as shown:
	//	buf[0] = buf[1] = middle 8 bits of Sample 1
	//	buf[1] = buf[0] =    top 8 bits of Sample 1
	//	buf[2] = buf[3]	= middle 8 bits of Sample 2
	//	buf[3] = buf[2] =    top 8 bits of Sample 2

		align 16
loop1:	sub		ecx, 5
		jl		done

		mov		eax, [esi]		; 3 2 1 0
		add		esi, 5
		bswap	eax				; 0 1 2 3
		ror		eax, 16			; 2 3 0 1
		mov		[edi], eax
		add		edi, 4

		jmp		loop1
done:
	}
}

inline void AudioSourcePCM::LPCM24M_to_WAVE16M(void *src, int len) {
    __asm {
		mov		esi, src
		mov		ecx, len
		mov		edi, esi

	// This is how the LPCM stream is coded:
	//	buf[0] =    top 8 bits of Sample 1
	//  buf[1] = middle 8 bits of Sample 1
	//	buf[2] =    top 8 bits of Sample 2
	//	buf[3] = middle 8 bits of Sample 2
	//	buf[4] = bottom 8 bits of Sample 1
	//  buf[5] = bottom 8 bits of Sample 2

	// We convert to Intel 16-bit as shown:
	//	buf[0] = buf[1] = middle 8 bits of Sample 1
	//	buf[1] = buf[0] =    top 8 bits of Sample 1
	//	buf[2] = buf[3]	= middle 8 bits of Sample 2
	//	buf[3] = buf[2] =    top 8 bits of Sample 2

		align 16
loop1:	sub		ecx, 6
		jl		done

		mov		eax, [esi]		; 3 2 1 0
		add		esi, 6
		bswap	eax				; 0 1 2 3
		ror		eax, 16			; 2 3 0 1
		mov		[edi], eax
		add		edi, 4

		jmp		loop1
done:
	}
}

inline void AudioSourcePCM::LPCM24S_to_WAVE16S(void *src, int len) {
    __asm {
		mov		esi, src
		mov		ecx, len
		mov		edi, esi

	// This is how the LPCM stream is coded:
	//	buf[0] =    top 8 bits of Left sample 1
	//  buf[1] = middle 8 bits of Left sample 1
	//	buf[2] =    top 8 bits of Right sample 1
	//	buf[3] = middle 8 bits of Right sample 1
	//	buf[4] =    top 8 bits of Left sample 2
	//  buf[5] = middle 8 bits of Left sample 2
	//	buf[6] =    top 8 bits of Right sample 2
	//	buf[7] = middle 8 bits of Right sample 2
	//	buf[8] = bottom 8 bits of Left sample 1
	//  buf[9] = bottom 8 bits of Right sample 1
	//	buf[A] = bottom 8 bits of Left sample 2
	//	buf[B] = bottom 8 bits of Right sample 2

	// We convert to Intel 16-bit as shown:
	//	buf[0] = buf[1] = middle 8 bits of Left sample 1
	//	buf[1] = buf[0] =    top 8 bits of Left sample 1
	//	buf[2] = buf[3]	= middle 8 bits of Right sample 1
	//	buf[3] = buf[2] =    top 8 bits of Right sample 1
	//	buf[4] = buf[5] = middle 8 bits of Left sample 2
	//	buf[5] = buf[4] =    top 8 bits of Left sample 2
	//	buf[6] = buf[7] = middle 8 bits of Right sample 2
	//	buf[7] = buf[6] =    top 8 bits of Right sample 2

		align 16
loop1:	sub		ecx, 6
		jl		done

		mov		eax, [esi]		; 3 2 1 0
		add		esi, 4
		bswap	eax				; 0 1 2 3
		ror		eax, 16			; 2 3 0 1
		mov		[edi], eax
		add		edi, 4

		sub		ecx, 6
		jl		done

		mov		eax, [esi]		; 7 6 5 4
		add		esi, 8
		bswap	eax				; 4 5 6 7
		ror		eax, 16			; 6 7 4 5
		mov		[edi], eax
		add		edi, 4

		jmp		loop1
done:
	}
}

int AudioSourcePCM::_read(VDPosition lStart, uint32 lCount, void *lpBuffer, uint32 cbBuffer, uint32 *lBytesRead, uint32 *lSamplesRead) {
	long lAudioPacket;
	MPEGSampleInfo *msi;
	long len,samples;

	long ba = getWaveFormat()->nBlockAlign;

	lAudioPacket = (long)lStart / PCM_BUFFER_SIZE;
	samples = PCM_BUFFER_SIZE - ((long)lStart % PCM_BUFFER_SIZE);

	if (lCount != AVISTREAMREAD_CONVENIENT)
		if (samples > lCount) samples = lCount;

	if (samples * ba > cbBuffer && lpBuffer) {
		samples = cbBuffer / ba;

		if (samples <= 0) {
			if (lSamplesRead) *lSamplesRead = 0;
			if (lBytesRead) *lBytesRead = 0;
			return AVIERR_BUFFERTOOSMALL;
		}
	}

	if (!lpBuffer) {
		if (lSamplesRead) *lSamplesRead = samples;
		if (lBytesRead) *lBytesRead = samples * ba;
		return AVIERR_OK;
	}

	if (lCurrentPacket != lAudioPacket) {
		msi = &parentPtr->audio_sample_list[lAudioPacket];
		len = msi->size;

		parentPtr->ReadStream(pkt_buffer, msi->stream_pos, len, true);

	// audio_first_header is:
	// 11111111 ........ ........ ........	non-zero
	// ........ x....... ........ ........	emphasis
	// ........ .x...... ........ ........	mute
	// ........ ..x..... ........ ........	reserved
	// ........ ...xxxxx ........ ........	frame number % 20
	// ........ ........ xxxxxxxx ........	dynamic range control
	// ........ ........ ........ xx......	00=16-,01=20-,10=24-bit (11=forbidden)
	// ........ ........ ........ ..xx....	00=48,01=96 KHz (others reserved?)
	// ........ ........ ........ ....x...	reserved
	// ........ ........ ........ .....xxx	number of channels - 1

		switch ((parentPtr->audio_first_header >> 6) & 3) {

		case 0:	// 16-bit LPCM
			LPCM16_to_WAVE16(pkt_buffer, len);
			break;

		case 1:	// 20-bit LPCM
			if (parentPtr->audio_first_header & 1) {
				LPCM20S_to_WAVE16S(pkt_buffer, len);
			} else {
				LPCM20M_to_WAVE16M(pkt_buffer, len);
			}
			break;

		case 2:	// 24-bit LPCM
			if (parentPtr->audio_first_header & 1) {
				LPCM24S_to_WAVE16S(pkt_buffer, len);
			} else {
				LPCM24M_to_WAVE16M(pkt_buffer, len);
			}
			break;
		}
		lCurrentPacket = lAudioPacket;
	}

	memcpy(lpBuffer, (short *)pkt_buffer + (lStart % PCM_BUFFER_SIZE) * (ba / 2), samples * ba);

	if (lSamplesRead) *lSamplesRead = samples;
	if (lBytesRead) *lBytesRead = samples*ba;

	return AVIERR_OK;
}

//////////////////////////////////////////////////////////////////////////
//
//
//					MPEGAudioParser
//
//
//////////////////////////////////////////////////////////////////////////

class MPEGAudioParser {
protected:
	unsigned long lFirstHeader;
	MPEGSampleInfo msi;
	__int64 bytepos;
	int hstate, skip;
private:
	unsigned long header;

public:
	MPEGAudioParser();

	virtual void Parse(const void *, int, DataVector *);
	virtual unsigned long getHeader();
};

MPEGAudioParser::MPEGAudioParser() {
	lFirstHeader = 0;
	header = 0;
	hstate = 0;
	skip = 0;
	bytepos = 0;
}

unsigned long MPEGAudioParser::getHeader() {
	return lFirstHeader;
}

void MPEGAudioParser::Parse(const void *pData, int len, DataVector *dst) {
	unsigned char *src = (unsigned char *)pData;

	while(len>0) {
		if (skip) {
			int tc = skip;

			if (tc > len)
				tc = len;

			len -= tc;
			skip -= tc;
			src += tc;

			// Audio frame finished?

			if (!skip) {
#ifdef STORE_TIME_STAMPS
				msi.PTS = -1;
#endif
				dst->Add(&msi);
			}

			continue;
		}

		// Collect header bytes.

		++hstate;
		header = (header>>8) | ((long)*src++ << 24);
		--len;

		MPEGAudioHeader hdr(header);
		if (hstate>=4 && hdr.IsValid()) {

			if (lFirstHeader && !hdr.IsConsistent(lFirstHeader))
				continue;

			// Okay, we like the header.

			hstate = 0;

			// Must be a frame start.

			if (!lFirstHeader)
				lFirstHeader = header;

			// Setup the sample information.  Don't add the sample, in case it's incomplete.

			msi.stream_pos	= bytepos + (src - (unsigned char *)pData) - 4;
			msi.header		= header;
			msi.size		= hdr.GetFrameSize();

			// Now skip the remainder of the sample.

			skip = msi.size-4;
		}
	}

	bytepos += src - (unsigned char *)pData;
}

//////////////////////////////////////////////////////////////////////////
//
//
//					AC3AudioParser
//
//
//////////////////////////////////////////////////////////////////////////

class AC3AudioParser : public MPEGAudioParser {
private:
	unsigned long syncinfo[2];
public:
	AC3AudioParser();
	void Parse(const void *, int, DataVector *);
};

AC3AudioParser::AC3AudioParser() {
	lFirstHeader = 0;
	hstate = 0;
	skip = 0;
	bytepos = 0;
	syncinfo[0] = 0;
}

void AC3AudioParser::Parse(const void *pData, int len, DataVector *dst) {
	unsigned char *src = (unsigned char *)pData;

	while(len>0) {
		if (skip) {
			int tc = skip;
			if (tc > len) tc = len;

			len -= tc;
			skip -= tc;
			src += tc;

			// Audio frame finished?
			if (!skip) {
#ifdef STORE_TIME_STAMPS
				msi.PTS = -1;
#endif
				dst->Add(&msi);
			}
			continue;
		}

		// Collect header bytes.
		if (syncinfo[0] == 0x770B) {

// 0B 77 mm ll fs va cm ??
//		0B 77	= syncword
//		mm ll	= MSB, LSB of CRC (ignore)
//		fs		= fscode, srcode
//		va		= version, audio service
//		cm		= audio coding mode
// store audio_first_header as [?? cm va fs]

			syncinfo[1] = (syncinfo[1] >> 8) | ((*src++) << 24);
			--len;
			if (++hstate == 6) {
				hstate = 0;

				if (skip = ac3_syncinfo(syncinfo, NULL, NULL, NULL)) {
					if (!lFirstHeader) lFirstHeader = syncinfo[1];
					msi.stream_pos = bytepos + (src - (unsigned char *)pData) - 8;
					msi.header = syncinfo[1];
					msi.size = skip;
					skip -= 8;
				}

				syncinfo[0] = 0;
			}
		} else {
			syncinfo[0] = (syncinfo[0] >> 8) | ((*src++) << 8);
			--len;
		}
	}

	bytepos += src - (unsigned char *)pData;
}

//////////////////////////////////////////////////////////////////////////
//
//
//					PCMAudioParser
//
//
//////////////////////////////////////////////////////////////////////////

class PCMAudioParser : public MPEGAudioParser {
public:
	PCMAudioParser();
	void Parse(const void *pData, int, DataVector *);
};

PCMAudioParser::PCMAudioParser() {
	lFirstHeader = 0;
	bytepos = 0;
	skip = 0;
}

void PCMAudioParser::Parse(const void *pData, int len, DataVector *dst) {

	if (pData) {

	// pData = LPCM header (6 bytes):
	//	0:	xxxxxxxx	number of frames
	//	1:	xxxxxxxx	offset of first frame (MSB)
	//	2:	xxxxxxxx	offset of first frame (LSB)
	//	3:	x.......	emphasis
	//		.x......	mute
	//		..x.....	reserved
	//		...xxxxx	frame number % 20
	//	4:	xx......	00=16-,01=20-,10=24-bit (11=forbidden)
	//		..xx....	00=48,01=96 KHz (others reserved?)
	//		....x...	reserved
	//		.....xxx	number of channels - 1
	//	5:	xxxxxxxx	dynamic range control

		unsigned char *src = (unsigned char *)pData;
		lFirstHeader = (src[3] << 8) + src[4] + (src[5] << 16);
		lFirstHeader |= 0xFF000000;	// ensure a non-zero lFirstHeader

	// lFirstHeader is:
	// 11111111 ........ ........ ........	non-zero
	// ........ x....... ........ ........	emphasis
	// ........ .x...... ........ ........	mute
	// ........ ..x..... ........ ........	reserved
	// ........ ...xxxxx ........ ........	frame number % 20
	// ........ ........ xxxxxxxx ........	dynamic range control
	// ........ ........ ........ xx......	00=16-,01=20-,10=24-bit (11=forbidden)
	// ........ ........ ........ ..xx....	00=48,01=96 KHz (others reserved?)
	// ........ ........ ........ ....x...	reserved
	// ........ ........ ........ .....xxx	number of channels - 1

		const int chans = (lFirstHeader & 7) + 1;
		const int bits = 16 + ((lFirstHeader & 0xC0) >> 4);

		//skip = chans * PCM_BUFFER_SIZE * (bits / 8);
		skip = chans * (PCM_BUFFER_SIZE / 8) * bits;

		msi.stream_pos = 0;
		msi.size = 0;
		return;
	}

	// Sadly this code will not actually parse all of the
	// audio data, but at most we may lose PCM_BUFFER_SIZE
	// (1536) samples.  At 48K this is 0.032 seconds.

	msi.size += len;

	while (msi.size >= skip) {
		int tc = msi.size - skip;

		msi.stream_pos = bytepos;
		msi.size = skip;
#ifdef STORE_TIME_STAMPS
		msi.PTS = -1;
#endif
		dst->Add(&msi);

		bytepos += skip;
		msi.size = tc;
	}
}

//////////////////////////////////////////////////////////////////////////
//
//
//					MPEGVideoParser
//
//
//////////////////////////////////////////////////////////////////////////

class MPEGVideoParser {
private:
	unsigned char buf[8];

	int idx, bytes;

	MPEGSampleInfo msi;
	__int64 bytepos;
	__int64 last_access_unit;

	long header;

	bool fPicturePending;
	bool fFoundSequenceStart;

public:
	VDFraction mFrameRate;
	int width, height;
	long seq_ext;
	int aspect_ratio_info;
	bool progressive_sequence;

	MPEGVideoParser();

	void setPos(__int64);
	void Parse(const void *, int, DataVector *);
};

MPEGVideoParser::MPEGVideoParser() {
	bytepos = 0;
	header = -1;

	fPicturePending = false;
	fFoundSequenceStart = false;
	seq_ext = 0;

	last_access_unit = 0;

	idx = bytes = 0;
	msi.frame_type = 0;
}

void MPEGVideoParser::setPos(__int64 pos) {
	bytepos = pos;
}

void MPEGVideoParser::Parse(const void *pData, int len, DataVector *dst) {
	unsigned char *src = (unsigned char *)pData;

	// bytes is zero when the class is created,
	// thus (idx<bytes) is false the first time
	while(len > 0) {

		if (idx < bytes) {
			int tc = bytes - idx;
			if (tc > len) tc = len;

			memcpy(buf+idx, src, tc);

			len -= tc;
			idx += tc;
			src += tc;

			if (idx >= bytes) {   // Finished?

				switch(header) {

				case VIDPKT_TYPE_PICTURE_START:
					msi.frame_type = (char)((buf[1] >> 3) & 7);
					msi.picture_structure = FRAME_PICTURE;
					msi.bitflags = PICTUREFLAG_PF;
					msi.fields = 2;
					fPicturePending = true;
					header = 0xFFFFFFFF;
					break;

				case VIDPKT_TYPE_SEQUENCE_START:
					//	12 bits: width
					//	12 bits: height
					//	 4 bits: aspect ratio
					//	 4 bits: picture rate
					if (bytes == 8) {
						width	= (buf[0] << 4) + (buf[1] >> 4);
						height	= ((buf[1] & 15) << 8) + buf[2];

						int frameratecode = buf[3] & 15;
						aspect_ratio_info = (buf[3] >> 4) & 15;

						switch(frameratecode) {
						case 1:	mFrameRate = VDFraction(24000, 1001); break;	// 1 (23.976) - NTSC FILM interlaced
						case 2:	mFrameRate = VDFraction(24000, 1000); break;	// 2 (24.000) - FILM
						case 3:	mFrameRate = VDFraction(25000, 1000); break;	// 3 (25.000) - PAL interlaced
						case 4:	mFrameRate = VDFraction(30000, 1001); break;	// 4 (29.970) - NTSC color interlaced
						case 5:	mFrameRate = VDFraction(30000, 1000); break;	// 5 (30.000) - NTSC b&w progressive
						case 6:	mFrameRate = VDFraction(50000, 1000); break;	// 6 (50.000) - PAL progressive
						case 7:	mFrameRate = VDFraction(60000, 1001); break;	// 7 (59.940) - NTSC color progressive
						case 8:	mFrameRate = VDFraction(60000, 1000); break;	// 8 (60.000) - NTSC b&w progressive
						default:
							//throw MyError("MPEG video stream contains an invalid frame rate (%d).", buf[3] & 15);
							VDLogAppMessage(kVDLogWarning, kVDST_Mpeg, kVDM_InvalidFramerate, 1, &frameratecode);
							mFrameRate = VDFraction(15, 1);	break;
						}

						frameratecode = (frameratecode - 1) & 7;
					}

					header = 0xFFFFFFFF;
					break;

				case VIDPKT_TYPE_EXT_START:

					if (bytes == 1) {
						int ext = buf[0] >> 4;

						if (ext == SEQUENCE_EXTENSION_ID
							&& fFoundSequenceStart
							&& seq_ext == 0) {

							bytes = 4;
							break;
						}

						if (ext == PICTURE_CODING_EXTENSION_ID
							&& fPicturePending) {

							bytes = 5;
							break;
						}

					} else if (bytes == 4) {
						memcpy(&seq_ext, buf, 4);

	// xxxx.... ........ ........	SEQUENCE_EXTENSION_ID
	// ....x... ........ ........	escape_bit
	// .....xxx xxxx.... ........	MPEG-2 profile / level
	// ........ ....x... ........	progressive_sequence
	// ........ .....xx. ........	chroma_format
	// ........ .......x x.......	horizontal_size_extension
	// ........ ........ .xx.....	vertical_size_extension

						progressive_sequence = (buf[1] >> 3) & 1;
						width = (width & 0xFFF) | ((buf[1] & 1) << 13) | ((buf[2] & 0x80) << 5);
						height = (height & 0xFFF) | ((buf[2] & 0x60) << 7);
						// frame rate extensions are not allowed
						// in any defined MPEG-2 profile

					} else if (bytes == 5) {

// xxxx.... ........ ........ ........ ........	PICTURE_CODING_EXTENSION_ID
// ....xxxx xxxxxxxx xxxx.... ........ ........	f_codes
// ........ ........ ....xx.. ........ ........	intra_dc_precision
// ........ ........ ......xx ........ ........	picture_structure
// ........ ........ ........ x....... ........	top_field_first
// ........ ........ ........ .x...... ........	frame_pred_frame_dct
// ........ ........ ........ ..x..... ........	concealment_vectors
// ........ ........ ........ ...x.... ........	q_scale_type
// ........ ........ ........ ....x... ........	intra_vlc_format
// ........ ........ ........ .....x.. ........	alternate_scan
// ........ ........ ........ ......x. ........	repeat_first_field
// ........ ........ ........ .......x ........	chroma_420_type
// ........ ........ ........ ........ x.......	progressive_frame

						msi.picture_structure = (char)(buf[2] & 3);
						if (buf[3] & 0x80) msi.bitflags |= PICTUREFLAG_TFF;
						if (buf[3] & 0x02) msi.bitflags |= PICTUREFLAG_RFF;
						if (!(buf[4] & 0x80)) msi.bitflags &= ~PICTUREFLAG_PF;

						// this is overly complicated
						if (	(msi.bitflags & PICTUREFLAG_RFF)
							&&	(msi.picture_structure == 3)
							&&	(msi.bitflags & PICTUREFLAG_PF)) {

							if (this->progressive_sequence) {
								msi.fields = (char)((msi.bitflags & PICTUREFLAG_TFF)? 6: 4);
							} else {
								msi.fields = 3;
							}
						} else if (msi.picture_structure != FRAME_PICTURE) {
							msi.fields = 1;
						} else {
							msi.fields = 2;
						}

					}

					header = 0xFFFFFFFF;
					break;
				}
			}
			continue;
		}

		// Look for a valid MPEG header
		header = (header<<8) + *src++;
		len--;

		if ((header & 0xFFFFFF00) == 0x100) {

			if (fPicturePending
				&& (header == VIDPKT_TYPE_SEQUENCE_START
				|| header == VIDPKT_TYPE_GROUP_START
				|| header == VIDPKT_TYPE_PICTURE_START
				|| header == VIDPKT_TYPE_SEQUENCE_END)) {

				// Don't add pictures until we've parsed the sequence!
				// Fix for Shakira: don't add picture if it is corrupt

				switch (msi.frame_type) {
				case MPEG_FRAME_TYPE_I:
				case MPEG_FRAME_TYPE_P:
				case MPEG_FRAME_TYPE_B:
					msi.stream_pos = last_access_unit;
					msi.size = (int)((bytepos + (src - (unsigned char *)pData) - 4) - msi.stream_pos);
					if (header == VIDPKT_TYPE_SEQUENCE_END) msi.size += 4;
#ifdef STORE_TIME_STAMPS
					msi.PTS = -1;
#endif
					dst->Add(&msi);
					break;
				}

				fPicturePending = false;
				last_access_unit = bytepos + (src - (unsigned char *)pData) - 4;
			}

			switch(header) {

			case VIDPKT_TYPE_SEQUENCE_START:
				if (!fFoundSequenceStart) {
					fFoundSequenceStart = true;
					last_access_unit = bytepos + (src - (unsigned char *)pData) - 4;
					idx = 0;
					bytes = 8;
				}
				break;

			case VIDPKT_TYPE_PICTURE_START:
				if (fFoundSequenceStart) {
					idx = 0;
					bytes = 2;
				}
				break;
/*
			case VIDPKT_TYPE_GROUP_START:
				if (fFoundSequenceStart) {
					idx = 0;
					bytes = 4;
				}
				break;
*/
			case VIDPKT_TYPE_EXT_START:
				if (fFoundSequenceStart) {
					idx = 0;
					bytes = 1;
				}
				break;

			default:
				header = 0xFFFFFFFF;
				break;
			}
		}
	}

	bytepos += src - (unsigned char *)pData;
}

//////////////////////////////////////////////////////////////////////////
//
//
//							MPEGFileParser
//
//
//////////////////////////////////////////////////////////////////////////

extern HWND g_hWnd;


class MPEGFileParser {
private:
	InputFileMPEG *parentPtr;

	static const unsigned short MPEG_AUDIO_DIALOG[];
	static const char szME[];

	__int64 video_stream_pos;
	__int64 audio_stream_pos;

	sint64 last_dts[52];

	// Fix for M2M, search at least 3 MB
	enum {
		SCAN_BUFFER_SIZE    = 65536,
		SEARCH_SYSTEM_LIMIT = 1048576*3,
		SEARCH_AUDIO_LIMIT  = 256
	};

	DataVector *video_stream_blocks;
	DataVector *video_stream_samples;
	DataVector *audio_stream_blocks;
	DataVector *audio_stream_samples;

	unsigned char *packet_buffer;
	unsigned char *pScanBuffer;
	unsigned char *pScan;
	unsigned char *pScanLimit;
	__int64 i64ScanCpos;

	VDFile  *mFile;
	int vidstream;
	bool fIsVCD;
	bool fAbort;
	long lFirstCode;
	HWND hWndStatus;

	__int64 file_len;

	void Rewind();
	int NextStartCode();
	bool Fill();
	bool Skip(int bytes);

	int Read();
	int Read(void *, int);

	__int64 Tell();

	bool Verify();
	inline long ScanAudio();

	bool ParsePack(int stream_id);

	static INT_PTR CALLBACK ParseDialogProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
	static INT_PTR CALLBACK MpegAudioDlg(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);

public:
	MPEGFileParser(InputFileMPEG *pp);
	~MPEGFileParser();

	MPEGVideoParser *videoParser;
	MPEGAudioParser *audioParser;

#ifndef STORE_TIME_STAMPS
	__int64 vEarliestPTS;
	__int64 aEarliestPTS;
#endif

	VDFraction frame_rate;
	bool fIsMPEG2;
	bool fIsPartial;
	bool fInterleaved;
	bool fIsScrambled;
	bool progressive_sequence;
	int aspect_ratio_info;
	int substream;

	void Parse(VDFile *handle);
};


const char MPEGFileParser::szME[]="MPEG File Parser";

MPEGFileParser::MPEGFileParser(InputFileMPEG *pp) {
	parentPtr = pp;

	// Clear variables
	pScan = pScanLimit = pScanBuffer = NULL;
	video_stream_blocks = NULL;
	video_stream_samples = NULL;
	audio_stream_blocks = NULL;
	audio_stream_samples = NULL;
	video_stream_pos = 0;
	audio_stream_pos = 0;
	packet_buffer = NULL;
	i64ScanCpos = 0;
	videoParser = NULL;
	audioParser = NULL;
#ifndef STORE_TIME_STAMPS
	vEarliestPTS = 0x7FFFFFFFFFFFFFFF;
	aEarliestPTS = 0x7FFFFFFFFFFFFFFF;
#endif
	hWndStatus = NULL;
	vidstream = 0;
	fIsScrambled = false;
	fIsPartial = true;
	fIsVCD = false;
	fAbort = false;
	mFile = NULL;
	substream = 0;
}

MPEGFileParser::~MPEGFileParser() {
	if (hWndStatus) {
		EnableWindow(GetParent(hWndStatus), TRUE);
		DestroyWindow(hWndStatus);
	}
	delete videoParser;
	delete audio_stream_samples;
	delete audio_stream_blocks;
	delete video_stream_samples;
	delete video_stream_blocks;
	delete audioParser;
	delete [] packet_buffer;
	delete [] pScanBuffer;
}

void MPEGFileParser::Rewind() {
	pScan = pScanLimit = pScanBuffer;
	i64ScanCpos = (fIsVCD)? 40: 0;
	mFile->seek(i64ScanCpos);
}

int MPEGFileParser::NextStartCode() {
	int i;

	do {
		if ((i = Read()) == 0) {
			if ((i = Read()) == 0) {
				do {
					i = Read();
				} while (i == 0);
				if (i == 1) {
					if ((i = Read()) != EOF) {
						return (i | 0x100);
					}
				}
			}
		}
	} while (i != EOF);

	return 0;
}

bool MPEGFileParser::Fill() {
	// If EOF, pScanLimit == pScanBuffer
	// If disk error return false
	int actual;

	i64ScanCpos += (pScanLimit - pScanBuffer);
	pScan = pScanLimit = pScanBuffer;

	if (fIsVCD) {
		actual = mFile->readData(pScanBuffer, 20);
		if (actual >= 0) {
			if (actual >= 20) {
				i64ScanCpos += 20;
				actual = mFile->readData(pScanBuffer, 2332);
			} else actual = 0;
		}
	} else {
		actual = mFile->readData(pScanBuffer, SCAN_BUFFER_SIZE);
	}
	if (actual < 0) return false;

	pScanLimit = pScanBuffer + actual;
	return true;
}

bool MPEGFileParser::Skip(int bytes) {
	if (bytes > 0) {
		int tc = pScanLimit - pScan;
		do {
			if (tc >= bytes) {
				pScan += bytes;
				bytes = 0;
				break;
			}
			bytes -= tc;
			if (!Fill()) break;
			tc = pScanLimit - pScanBuffer;
		} while (tc > 0);
	}
	return (bytes <= 0);
}

int MPEGFileParser::Read() {
	if (pScan < pScanLimit) return *pScan++;
	unsigned char c;
	if (Read(&c, 1) == 1) return c;
	return EOF;
}

int MPEGFileParser::Read(void *buffer, int bytes) {
	int tc, actual;
	int total = 0;

	if (bytes > 0) {

		// Start by emptying the buffer
		if (pScan < pScanLimit) {
			tc = pScanLimit - pScan;
			if (tc >= bytes) {
				memcpy(buffer, pScan, bytes);
				pScan += bytes;
				return bytes;
			}
			memcpy(buffer, pScan, tc);
			buffer = (char *)buffer + tc;
			bytes -= tc;
			total = tc;
		}

		// Now:
		//  bytes is greater than zero
		//  the scan buffer is empty
		i64ScanCpos += (pScanLimit - pScanBuffer);
		pScan = pScanLimit = pScanBuffer;

		// Read chunks directly until bytes <= chunksize
		tc = (fIsVCD)? 2332: SCAN_BUFFER_SIZE;

		while (bytes > tc) {
			if (fIsVCD) {
				actual = mFile->readData(pScanBuffer, 20);
				if (actual < 20) {
					if (actual < 0)
						throw MyError("%s: read error", szME);
					return total;
				}
				i64ScanCpos += 20;
			}
			actual = mFile->readData(buffer, tc);
			if (actual < tc) {
				if (actual < 0)
					throw MyError("%s: read error", szME);
				pScan = pScanLimit = (pScanBuffer + actual);
				return (total + actual);
			}
			buffer = (char *)buffer + tc;
			i64ScanCpos += tc;
			total += tc;
			bytes -= tc;
		}

		// Now, bytes <= chunksize
		if (fIsVCD) {
			actual = mFile->readData(pScanBuffer, 20);
			if (actual < 20) {
				if (actual < 0)
					throw MyError("%s: read error", szME);
				return total;
			}
			i64ScanCpos += 20;
		}
		actual = mFile->readData(pScanBuffer, tc);
		if (actual < 0)
			throw MyError("%s: read error", szME);

		if (actual > 0) {
			pScanLimit = (pScanBuffer + actual);
			if (actual > bytes) actual = bytes;
			memcpy(buffer, pScanBuffer, actual);
			pScan += actual;
			total += actual;
		}
	}
	return total;
}

__int64 MPEGFileParser::Tell() {
	return i64ScanCpos + (pScan - pScanBuffer);
}

// Create MPEG audio dialog in memory
const unsigned short MPEGFileParser::MPEG_AUDIO_DIALOG[] = {
	// DLGTEMPLATE
	0x0040, 0x00C0,		// style (DS_SETFONT | WS_CAPTION)
	0, 0,				// dwExtendedStyle
	3,					// cdit
	0, 0, 256, 96,		// x, y, cx, cy
	0, 0,				// menu, class
	'M','u','l','t','i','p','l','e',' ','A','u','d','i','o',' ',
	'S','t','r','e','a','m','s',' ','F','o','u','n','d',' ',0,
	8,					// pointsize
	'M','S',' ','S','a','n','s',' ','S','e','r','i','f',0,

	// DLGITEMTEMPLATE
	0x0001, 0x5002,		// style (WS_CHILD | WS_VISIBLE | WS_GROUP | SS_CENTER)
	0, 0,				// dwExtendedStyle
	80, 4, 96, 8,		// x, y, cx, cy
	-1,					// id (IDC_STATIC)
	0xFFFF, 0x0082,		// class (Static)
	'C','h','o','o','s','e',' ','a','n',' ',
	'a','u','d','i','o',' ','s','t','r','e','a','m',':',0,
	0,					// data

	// DLGITEMTEMPLATE
	0x0101, 0x50A1,		// style (WS_CHILD | WS_VISIBLE | WS_BORDER |
	// WS_VSCROLL | WS_TABSTOP | LBS_NOINTEGRALHEIGHT | LBS_NOTIFY)
	0, 0,				// dwExtendedStyle
	4, 16, 248, 56,		// x, y, cx, cy
	IDOK+1,				// id
	0xFFFF, 0x0083,		// class (List Box)
	'L','i','s','t','1',0,
	0,					// data

	// DLGITEMTEMPLATE
	0x0001, 0x5001,		// style (WS_CHILD | WS_VISIBLE | WS_TABSTOP | BS_DEFPUSHBUTTON)
	0, 0,				// dwExtendedStyle
	100, 76, 56, 16,	// x, y, cx, cy
	IDOK,				// id
	0xFFFF, 0x0080,		// class (Button)
	'O','K',0,
	0					// data
};

INT_PTR CALLBACK MPEGFileParser::MpegAudioDlg(HWND hDlg, UINT uMsg,
	WPARAM wParam, LPARAM lParam) {

	switch (uMsg) {

	case WM_INITDIALOG:
		{
			TCHAR buf[64];
			__int64 audiomask = *((__int64 *)lParam);

			HWND hwndList = GetDlgItem(hDlg, IDOK+1);
			int i;

			// Any MPEG audio streams?
			for (i = 0; i < 32; i++) {
				if ((audiomask >> i) & 1) {
					wsprintf(buf, "MPEG Audio Stream 0x%X", 0xC0+i);
					SendMessage(hwndList, LB_ADDSTRING, 0, (LPARAM)buf);
				}
			}

			// Any AC-3 audio streams?
			for (i = 32; i < 40; i++) {
				if ((audiomask >> i) & 1) {
					wsprintf(buf, "Dolby AC-3 Audio (substream 0x8%X)", i-32);
					SendMessage(hwndList, LB_ADDSTRING, 0, (LPARAM)buf);
				}
			}

			// Any LPCM audio streams?
			for (i = 40; i < 48; i++) {
				if ((audiomask >> i) & 1) {
					wsprintf(buf, "Linear PCM Audio (substream 0xA%X)", i-40);
					SendMessage(hwndList, LB_ADDSTRING, 0, (LPARAM)buf);
				}
			}

			if (audiomask)
				SendMessage(hwndList,LB_SETCURSEL,0,0);

			return TRUE;
		}

	case WM_COMMAND:
		if (LOWORD(wParam) == IDOK) {
			HWND hwndList = GetDlgItem(hDlg, IDOK+1);
			int i = SendMessage(hwndList, LB_GETCURSEL, 0, 0);
			EndDialog(hDlg, i);
			return TRUE;
		}
		break;
	}

	return FALSE;
}

bool MPEGFileParser::Verify() {

	// Check the validity of an MPEG file

    long hdr, count;
	unsigned char buf[4];
	long lFirstPES;
	int i;

	fIsVCD = false;
	fIsMPEG2 = false;
	fInterleaved = false;
	lFirstCode = -1;
	lFirstPES = -1;

	// Search for system layer
	// Emotion.m2v has empty PES headers (0x1E0) but no system!

	Rewind();
	Read(buf, 4);

	if (buf[0] == 'R' && buf[1] == 'I' && buf[2] == 'F' && buf[3] == 'F') {
		Skip(4);
		Read(buf, 4);
		if (buf[0] == 'C' && buf[1] == 'D' && buf[2] == 'X' && buf[3] == 'A') {
			fIsVCD = true;
		}
	}

	// If this is a VCD/SVCD, setting fIsVCD true and calling
	// Rewind() resets the reader so it reads only data blocks

	Rewind();
	Read(buf, 4);
	hdr = (buf[0] << 24) + (buf[1] << 16) + (buf[2] << 8) + buf[3];
	count = 4;

	while (count < SEARCH_SYSTEM_LIMIT) {

		if (hdr == 0x1BA) {

			// Verify pack header
			// (This is very robust)

			Read(buf, 4);
			count += 4;

			if (	(buf[0] & 0xC4) == 0x44
				&&	(buf[2] & 4) == 4) {

			// MPEG-2 pack
			// 01sss1ss ssssssss sssss1ss ssssssss sssss1ee
			// eeeeeee1 rrrrrrrr rrrrrrrr rrrrrr11 11111xxx

			// If we fail, buf[3] might still begin
			// a start code, so don't toss it yet!

				buf[0] = buf[3];

				Read(buf+1, 3);
				count += 3;

				if (	(buf[1] & 4) == 4
					&&	(buf[2] & 1) == 1) {

					// Almost there!

					buf[0] = buf[3];

					Read(buf+1, 3);
					count += 3;

					const long mux	= (buf[0] << 14)
									+ (buf[1] << 6)
									+ (buf[2] >> 2);

					if (mux && (buf[2] & 3) == 3) {
						// Yes!!

						Skip(buf[3] & 7);
						fInterleaved = true;
						fIsMPEG2 = true;
						lFirstCode = count - 14;
						break;
					}
				}

			} else if (	(buf[0] & 0xF1) == 0x21
					&&	(buf[2] & 1) == 1) {

			// MPEG-1 pack
			// 0010aaa1 aaaaaaaa aaaaaaa1 aaaaaaaa
			// aaaaaaa1 1mmmmmmm mmmmmmmm mmmmmmm1

				buf[0] = buf[3];

				Read(buf+1, 3);
				count += 3;

				if (	(buf[1] & 1) == 1
					&&	(buf[2] & 0x80) == 0x80) {

					// Almost there!

					buf[0] = buf[1];
					buf[1] = buf[2];
					buf[2] = buf[3];
					buf[3] = (unsigned char)Read();
					count++;

					const long mux	= (buf[1] << 15)
									+ (buf[2] << 7)
									+ (buf[3] >> 1);

					if (mux && (buf[3] & 1) == 1) {
						// Yes!!

						fInterleaved = true;
						lFirstCode = count - 12;
						break;
					}
				}
			}

			// Damn, we were fooled
			// Back up four bytes

			hdr = (buf[0] << 24) + (buf[1] << 16) + (buf[2] << 8) + buf[3];
			count -= 4;
			continue;
		}

/*
		if (hdr == 0x1BB) {
			// Verify system header

			Read(buf, 4);
			count += 4;

			const long rb	= ((buf[0] & 0x7F) << 15)
							+ (buf[1] << 7)
							+ (buf[2] >> 1);

		// 1....... ........ ........ ........	marker_bit
		// .xxxxxxx xxxxxxxx xxxxxxx. ........	rate_bound
		// ........ ........ .......1 ........	marker_bit
		// ........ ........ ........ xxxxxx..	audio_bound
		// ........ ........ ........ ......x.	fixed_flag
		// ........ ........ ........ .......x	CSPS_flag

			if ((buf[0] & 0x80) == 0x80 && (buf[2] & 1) && rb) {
				// We can discard the first 3 bytes
				buf[0] = buf[3];
				Read(buf+1, 2);
				count += 2;

		// x....... ........	system_audio_lock_flag
		// .x...... ........	system_video_lock_flag
		// ..1..... ........	marker_bit
		// ...xxxxx ........	video_bound
		// ........ xxxxxxxx	reserved

				if ((buf[1] & 0x20) == 0x20) {
					// Yes!!

					fInterleaved = true;
					lFirstCode = count - 10;
					break;
				}
			}

			// Damn, we were fooled
			// Back up four bytes

			hdr = (buf[0] << 24) + (buf[1] << 16) + (buf[2] << 8) + buf[3];
			count -= 4;
			continue;
		}
*/

		// If we hit a sequence header, remember it

		if (hdr == VIDPKT_TYPE_SEQUENCE_START) {

			Read(buf, 4);
			count += 4;

			// wwwwwwww wwwwhhhh hhhhhhhh aaaarrrr
			// bbbbbbbb bbbbbbbb bb1vvvvv vvvvvxx.

			const int w = (buf[0] << 4) + (buf[1] >> 4);
			const int h = ((buf[1] & 15) << 4) + buf[2];

			if (	w && h
				&&	(buf[3] >> 4)
				&&	(buf[3] & 15)) {

				// In this case, we can discard the whole buffer
				// because no start code can begin at any byte

				Read(buf, 4);
				count += 4;

				const int b = (buf[0] << 10) + (buf[1] << 2) + (buf[2] >> 6);

				if (b && (buf[2] & 0x20) == 0x20) {
					// Yay!

					// Let's say now that IF we've found multiple
					// valid sequence headers, AND we've searched
					// ~70000 bytes for pack headers, then it's
					// fairly certain this is an elementary stream.

					if (lFirstCode != -1) {
						// We found at least one sequence
						// header prior to this one
						if (count >= 70000) break;

					} else {
						lFirstCode = count - 12;
					}
				}
			}

			// Damn, fooled again
			// Back up four bytes

			hdr = (buf[0] << 24) + (buf[1] << 16) + (buf[2] << 8) + buf[3];
			count -= 4;
			continue;
		}

/*
		if ((hdr >= 0x1C0 && hdr <= 0x1EF) || hdr == 0x1BD) {
			// Some TS to PS converts have multiplexed
			// streams but no pack headers!

			// Let's try this hack...
			// We'll accept a subsequent PES for verification
			// only if we have previously seen both a PES and
			// a video sequence header. In this case, the
			// earliest PES becomes "lFirstCode"...

			if (lFirstPES != -1) {
				if (lFirstCode != -1) {
					Read(buf, 3);
					fInterleaved = true;
					fIsMPEG2 = ((buf[2] & 0xC0) == 0x80);
					lFirstCode = lFirstPES;
					break;
				}
			} else {
				lFirstPES = count - 4;
			}
		}
*/
		if ((i = Read()) == EOF) break;
		hdr = (hdr << 8) + i;
		count++;
    }

    if (lFirstCode < 0) return false;

	if (!fInterleaved) {
		if (fIsVCD) return false;	// DOH!

		// No system layer...
		// Jump back to the sequence header,
		// and check for sequence extension

		Rewind();
		Skip(lFirstCode + 8);

		if (NextStartCode() == VIDPKT_TYPE_EXT_START) {
			fIsMPEG2 = ((Read() >> 4) == SEQUENCE_EXTENSION_ID);
		}
	}

	// Looks good to me
	return true;
}

inline long MPEGFileParser::ScanAudio() {

    long i, hdr, count;
	long packet_size;

	__int64 audiomask = 0;

	// audiomask[31...0] = MPEG Audio
	// audiomask[39..32] = AC3
	// audiomask[47..40] = LPCM

	Rewind();
	Skip(lFirstCode);

	// Search for audio packets
	// (This is not as robust as it could be...)

	for (count = 0; count < SEARCH_AUDIO_LIMIT; count++) {

		hdr = NextStartCode();

		if (hdr == 0x1BA) {

			if ((Read() & 0xC4) == 0x44) {
				Skip(8);
				i = Read();
				Skip(i & 7);

			} else
				Skip(7);

        } else {
			if (!hdr || hdr == 0x1B9) break;

			packet_size = Read();
			packet_size = (packet_size << 8) + Read();

			if ((i = Read()) == EOF) break;
			packet_size--;

			if (hdr == 0x1BD) {
				// private_stream_1
				// AC-3, LPCM, DTS, (SDDS?)

				if ((i & 0xC0) == 0x80) {
					// MPEG-2 PES

					if (i & 0x30) fIsScrambled = true;

					i = Read();	// PTS/DTS flags
					i = Read();	// PES_header_data_length

					Skip(i);
					packet_size -= (i + 2);

				} else {
					// MPEG-1 PES

					while (i == 0xFF) {
						if (--packet_size <= 0) break;
						i = Read();
					}

					if (i & 0x40) {
						i = Read();
						i = Read();
						packet_size -= 2;
					}

					if (i < 0x40) {
						if (i & 0x20) {
							if (i & 0x10) i = 9; else i = 4;
							packet_size -= i;
							Skip(i);
						}
					}
				}

				int ss = Read();	// substream
				if (ss == EOF) break;
				packet_size--;

				if ((ss & 0xF8) == 0x80) {

					// Probably AC-3, but let's not assume...
					// (DTS Audio has a syncword of 7FFF8001)
					// 1 byte = # of syncwords (access units) in packet
					// 2 bytes = 1-based offset of first syncword

					packet_size--;
					if (Read() > 0) {

						i = Read();
						i = (i << 8) + Read();
						packet_size -= 2;

		// Say packet_size is 7 here (# of bytes remaining in pack)
		//		### ohi olo xxx xxx xxx xxx xxx xxx xxx
		//		offset:      1   2   3   4   5   6   7
		// Thus "offset" must be greater than 0 and less than 7

						if (i > 0 && i < packet_size) {

							Skip(i - 1);
							packet_size -= i;

							if (Read() == 0x0B) {
								packet_size--;
								if (Read() == 0x77) {
									// Yay!
									audiomask |= (1i64 << (32 + (ss & 7)));
								}
							}
						}
					}

				} else if ((ss & 0xF8) == 0xA0) {

		// LPCM header:
		//	1 byte	= # of frames in packet
		//	2 bytes	= 1-based offset of first frame
		//	1 byte:	x.......	emphasis
		//			.x......	mute
		//			..x.....	reserved
		//			...xxxxx	frame number % 20
		//	1 byte:	xx......	00=16-,01=20-,10=24-bit (11=forbidden)
		//			..xx....	00=48,01=96 KHz (others reserved?)
		//			....x...	reserved
		//			.....xxx	number of channels - 1
		//	1 byte:	xxxxxxxx	dynamic range control

					packet_size--;
					if (Read() > 0) {

						i = Read();
						i = (i << 8) + Read();
						packet_size -= 2;

						if (i > 0 && i < packet_size) {
							i = Read();
							i = Read();
							packet_size -= 2;

		// We can support 48K or 96K 16- or 24-bit mono and stereo

		//	1 byte:	xx......	00=16-,01=20-,10=24-bit (11=forbidden)
		//			..xx....	00=48,01=96 KHz (others reserved?)
		//			....x...	reserved
		//			.....xxx	number of channels - 1

							if ((i & 0xC0) != 0xC0 && (i & 0x26) == 0) {
							//if ((i & 0x66) == 0) {
								// Yay!
								audiomask |= (1i64 << (40 + (ss & 7)));
							}
						}
					}
				}

			} else if (hdr >= 0x1C0 && hdr <= 0x1DF) {
				// MPEG Audio

				if ((i & 0xC0) == 0x80 && (i & 0x30) != 0)
					fIsScrambled = true;

				audiomask |= (1i64 << (hdr & 31));
			}

			if (packet_size > 0) Skip(packet_size);
		}
	}

	//_RPT1(0, "MPEGFileParser::ScanAudio: audiomask = 0x%I64X\n", audiomask);

	hdr = -1;
	count = 0;
	for (i = 0; i < 48; i++) {
		if ((audiomask >> i) & 1) {
			if (hdr == -1) hdr = i;
			count++;
		}
	}

	if (count == 0) return 0;	// no audio found

	// If there is only one audio stream, return that

	if (count == 1) {
		if (hdr >= 40) {
			hdr = 0xA0 + (hdr - 40);
		} else if (hdr >= 32) {
			hdr = 0x80 + (hdr - 32);
		} else {
			hdr = 0xC0 + hdr;
		}
		return hdr;
	}

	// Let the user select an audio stream

	hdr = DialogBoxIndirectParam(GetModuleHandle(NULL),
		(LPCDLGTEMPLATE)MPEG_AUDIO_DIALOG, g_hWnd,
		MpegAudioDlg, (LPARAM)&audiomask);

	count = 0;
	for (i = 0; i < 48; i++) {
		if ((audiomask >> i) & 1) {
			if (hdr == count) break;
			count++;
		}
	}

	if (i >= 48) {
		i = 0;
	} else if (i >= 40) {
		i = 0xA0 + (i - 40);
	} else if (i >= 32) {
		i = 0x80 + (i - 32);
	} else {
		i = 0xC0 + i;
	}
	return i;
}

bool MPEGFileParser::ParsePack(int stream_id) {
	// Parse one elementary stream packet header
	// Return "false" if packet is incomplete
	int i, actual, ptsflags;
	MPEGPacketInfo mpi;
	int pack_length = 0;
	sint64 pts, dts, tagpos;
	bool complete = false;

	switch (stream_id) {

	case 0x1BA:
		pack_length = 0;
		if ((i = Read()) != EOF) {
			if ((i & 0xC4) == 0x44) {
				// MPEG-2
				if (Skip(8)) {
					if ((i = Read()) != EOF) {
						complete = (Skip(i & 7));
					}
				}
			} else if ((i & 0xF1) == 0x21) {
				// MPEG-1
				complete = (Skip(7));
			}
		}
		break;

	default:
		tagpos = Tell();
#ifdef STORE_TIME_STAMPS
		mpi.PTS = -1;
#endif

		if (fInterleaved) {
			pack_length = Read();
			pack_length = (pack_length << 8) + Read();
			if (pack_length == 0) {
				complete = true;
				break;
			}

			if ((stream_id >= 0x1C0 && stream_id <= 0x1EF) || stream_id == 0x1BD) {
				// MPEG Audio, MPEG Video, or private_stream_1

				// parse stream header
				i = Read();
				--pack_length;

				if ((i & 0xC0) == 0x80) {

					// MPEG-2 packet
					fIsScrambled = (i & 0x30) != 0;
					ptsflags = Read();  // PTS/DTS flags
					i = Read();         // PES_header_data_length
					pack_length -= 2;

					if (ptsflags & 0x80) {  // PTS follows
						//  0010ppp1 pppppppp ppppppp1 pppppppp ppppppp1
						pts = (Read() & 0x0E) >> 1;
						pts = (pts << 8) + Read();
						pts = (pts << 7) + (Read() >> 1);
						pts = (pts << 8) + Read();
						pts = (pts << 7) + (Read() >> 1);
						pack_length -= 5;
						i -= 5;
#ifdef STORE_TIME_STAMPS
						mpi.PTS = pts;
#endif
						if (ptsflags & 0x40) {  // DTS follows
							// 0001ddd1 dddddddd ddddddd1 dddddddd ddddddd1
							dts = (Read() & 0x0E) >> 1;
							dts = (dts << 8) + Read();
							dts = (dts << 7) + (Read() >> 1);
							dts = (dts << 8) + Read();
							dts = (dts << 7) + (Read() >> 1);
							pack_length -= 5;
							i -= 5;
						} else dts = pts;
					}
					if (i > 0) {
						Skip(i);
						pack_length -= i;
					}

				} else {
					// MPEG-1 packet
					while (i == 0xFF) {
						if (--pack_length <= 0) break;
						i = Read();
					}
					if (pack_length <= 0) break;
					if (i & 0x40) {
						Skip(1);    // P-STD_buffer size
						i = Read();
						pack_length -= 2;
					}
					ptsflags = i << 2;
					if (ptsflags & 0x80) {  // PTS follows
						//  0010ppp1 pppppppp ppppppp1 pppppppp ppppppp1
						pts = (i & 0x0E) >> 1;
						pts = (pts << 8) + Read();
						pts = (pts << 7) + (Read() >> 1);
						pts = (pts << 8) + Read();
						pts = (pts << 7) + (Read() >> 1);
						pack_length -= 4;
#ifdef STORE_TIME_STAMPS
						mpi.PTS = pts;
#endif
						if (ptsflags & 0x40) {  // DTS follows
							// 0001ddd1 dddddddd ddddddd1 dddddddd ddddddd1
							dts = (Read() & 0x0E) >> 1;
							dts = (dts << 8) + Read();
							dts = (dts << 7) + (Read() >> 1);
							dts = (dts << 8) + Read();
							dts = (dts << 7) + (Read() >> 1);
							pack_length -= 5;
						} else dts = pts;
					}
				}

				if (ptsflags & 0x80) {

#ifndef STORE_TIME_STAMPS
					if ((stream_id & 0xF0) == 0xE0) {
						// Video
						if (pts < vEarliestPTS) vEarliestPTS = pts;
					} else {
						// Audio
						if (pts < aEarliestPTS) aEarliestPTS = pts;
					}
#endif
					sint64 last_stream_dts = last_dts[stream_id - 0x1BD];
					if (last_stream_dts != -1) {
						// decoding timestamps should never run backwards
						// and must occur at least every 0.7s (90KHz clock)
						if ((dts - last_stream_dts) > 63000) {
							const wchar_t *pStreamType = (stream_id < 0x1E0)? L"audio" : L"video";
							const int nStream = (stream_id - 0x1C0) & 0x1F;
							VDLogAppMessage(kVDLogWarning, kVDST_Mpeg, kVDM_OOOTimestamp, 5, &pStreamType, &nStream, &tagpos, &last_stream_dts, &dts);
						}
					}
					last_dts[stream_id - 0x1BD] = dts;
				}
            }
			if (pack_length <= 0) {
				complete = (pack_length == 0);
				break;
			}

		} else {
			// not interleaved
			stream_id = 0x1E0;
			pack_length = 65536;
		}
        
		if (stream_id == 0x1BD) {
			// private_stream_1, AC3 or LPCM
			i = Read();
			--pack_length;
			if ((i & 0xF8) == 0x80) {
				if (audioParser == NULL) {
					if (audioParser = new AC3AudioParser()) {
						substream = i;
					}
				}

				if (i == substream && audioParser) {
	// 1 byte  = # of syncwords (i.e., access units) in packet
	// 2 bytes = 1-based offset of first syncword (0x0B77)
					Skip(3);
					pack_length -= 3;

					mpi.file_pos = Tell();
					mpi.stream_pos = audio_stream_pos;
					audio_stream_blocks->Add(&mpi);

					actual = Read(packet_buffer, pack_length);
					if (actual >= pack_length) complete = true;
					audioParser->Parse(packet_buffer, actual, audio_stream_samples);
					audio_stream_pos += actual;
					pack_length = 0;
				}
			
			} else if ((i & 0xF8) == 0xA0) {

				unsigned char buf[6];
				int frame_offset;

				Read(buf, 6);
				pack_length -= 6;

				frame_offset = (buf[1] << 8) + buf[2];

	//	0:	xxxxxxxx	number of frames
	//	1:	xxxxxxxx	offset of first frame (MSB)
	//	2:	xxxxxxxx	offset of first frame (LSB)
	//	3:	x.......	emphasis
	//		.x......	mute
	//		..x.....	reserved
	//		...xxxxx	frame number % 20
	//	4:	xx......	00=16-,01=20-,10=24-bit (11=forbidden)
	//		..xx....	00=48,01=96 KHz (others reserved?)
	//		....x...	reserved
	//		.....xxx	number of channels - 1
	//	5:	xxxxxxxx	dynamic range control

	// We can support 48K or 96K 16- or 24-bit mono and stereo

				if (	buf[0] > 0
					&&	frame_offset > 0
					&&	frame_offset < pack_length
					&&	(buf[4] & 0xC0) != 0xC0
					&&	(buf[4] & 0x26) == 0) {

					// Looks good

					if (audioParser == NULL) {
						if (audioParser = new PCMAudioParser()) {
							substream = i;
						}
					}

					if (i == substream && audioParser) {

						if (audioParser->getHeader() == 0) {
							// Prime him with the first header
							audioParser->Parse(buf, 6, NULL);
						}

						mpi.file_pos = Tell();
						mpi.stream_pos = audio_stream_pos;
						audio_stream_blocks->Add(&mpi);

						actual = Read(packet_buffer, pack_length);
						if (actual >= pack_length) complete = true;

						audioParser->Parse(NULL, actual, audio_stream_samples);
						audio_stream_pos += actual;
						pack_length = 0;
					}
				}
			}

		} else if (stream_id >= 0x1E0 && stream_id <= 0x1EF) {
			// MPEG video (only the first stream found)
			if (!vidstream) vidstream = stream_id;
			if (stream_id == vidstream) {
				if (fInterleaved) {
					mpi.file_pos = Tell();
					mpi.stream_pos = video_stream_pos;
					video_stream_blocks->Add(&mpi);
				}
				actual = Read(packet_buffer, pack_length);
				if (actual >= pack_length) complete = true;
				videoParser->Parse(packet_buffer, actual, video_stream_samples);
				video_stream_pos += actual;
				pack_length = 0;
			}

		} else if (stream_id >= 0x1C0 && stream_id <= 0x1DF) {
			// MPEG audio
			i = (stream_id & 0xFF);
			if (audioParser == NULL) {
				if (audioParser = new MPEGAudioParser()) {
					substream = i;
				}
			}
			if (i == substream && audioParser) {
				mpi.file_pos = Tell();
				mpi.stream_pos = audio_stream_pos;
				audio_stream_blocks->Add(&mpi);

				actual = Read(packet_buffer, pack_length);
				if (actual >= pack_length) complete = true;
				audioParser->Parse(packet_buffer, actual, audio_stream_samples);
				audio_stream_pos += actual;
				pack_length = 0;
			}
		}
    }

	if (pack_length > 0) return Skip(pack_length);
	return complete;
}

void MPEGFileParser::Parse(VDFile *handle) {
	int i;

	if ((this->mFile = handle) == NULL)
		throw MyError("%s: invalid file handle", szME);

	// Determine the file's size
	file_len = mFile->size();

	if (file_len < 0)
		throw MyError("%s: couldn't determine file size", szME);

	pScanBuffer = new unsigned char[SCAN_BUFFER_SIZE];
	if (pScanBuffer == NULL) throw MyMemoryError();

	// Clear all DTS to -1 (an impossible number)
	memset(last_dts, 255, 52*sizeof(sint64));

	// Verify the MPEG file.  If successful,
	// This sets up the following variables:
	//		lFirstCode
	//		fInterleaved
	//		fIsMPEG2
	//		fIsVCD
	if (!Verify())
		throw MyError("%s: couldn't determine MPEG type", szME);


	if (fInterleaved) {

#ifdef REMEMBER_AUDIO
		if (parentPtr->iChosenAudio)
			substream = parentPtr->iChosenAudio;
		else
#endif
			substream = ScanAudio();


		switch (substream & AUDIOTYPE_MASK) {
		case AUDIOTYPE_AC3:
			audioParser = new AC3AudioParser();
			break;
		case AUDIOTYPE_LPCM:
			audioParser = new PCMAudioParser();
			break;
		case AUDIOTYPE_MPEG:
			audioParser = new MPEGAudioParser();
			break;
		}

	} else {

		// Beware of Transport Streams!
		// These will appear valid, however:
		//	- fInterleaved will be false
		//	- lFirstCode will be at least 4
		//	- The first byte will be 0x47

		if (lFirstCode >= 4) {
			Rewind();
			if (Read() == 0x47) {
				VDLogAppMessage(kVDLogWarning, kVDST_Mpeg, kVDM_TransportStream);
			}
		}
	}

	// Begin file parsing!  This is a royal bitch!

	try {
		packet_buffer = new unsigned char[65536];

		video_stream_blocks = new DataVector(sizeof MPEGPacketInfo);
		video_stream_samples = new DataVector(sizeof MPEGSampleInfo);
		video_stream_pos = 0;

		audio_stream_blocks = new DataVector(sizeof MPEGPacketInfo);
		audio_stream_samples = new DataVector(sizeof MPEGSampleInfo);
		audio_stream_pos = 0;

		videoParser = new MPEGVideoParser();

		bool end_of_file = false;

		try {
			Rewind();

			// pop up the dialog...
			if (!(hWndStatus = CreateDialogParam(GetModuleHandle(NULL),
				MAKEINTRESOURCE(IDD_PROGRESS), g_hWnd, ParseDialogProc, (LPARAM)this)))
					throw MyMemoryError();

			Skip(lFirstCode);

			do {
				if (!guiDlgMessageLoop(hWndStatus))
					fAbort = true;

				if (fAbort) break;

				if (fInterleaved)
					i = NextStartCode();
				else i = 0x1E0;

				switch (i) {

				case VIDPKT_TYPE_SEQUENCE_END:
					if (fInterleaved) break;
				case 0x1B9:     // ISO 11172 end code
					fIsPartial = false;
				case 0:         // no more start codes
					end_of_file = true;
					break;

				default:
					if (i > 0x1B9)
						end_of_file = !ParsePack(i);
				}
                
			} while (!end_of_file);

		} catch(const MyError&) {
			throw;
		}

		// If the last frame is incomplete, this start
		// code will prevent it from being included...

		i = ((VIDPKT_TYPE_USER_START & 0xFF) << 24) | 0x00010000;
		videoParser->Parse(&i, 4, video_stream_samples);

        // Construct stream and packet lookup tables, and
		// transfer them to the parent (I don't like this, but...)

        if (fInterleaved) {
			MPEGPacketInfo mpi;

			mpi.file_pos = 0;
			mpi.stream_pos = video_stream_pos;
			video_stream_blocks->Add(&mpi);

			parentPtr->video_packet_list = (MPEGPacketInfo *)
                video_stream_blocks->MakeArray();
            parentPtr->packets = video_stream_blocks->Length() - 1;

            if (audioParser) {
				mpi.file_pos = 0;
				mpi.stream_pos = audio_stream_pos;
				audio_stream_blocks->Add(&mpi);

                parentPtr->audio_packet_list = (MPEGPacketInfo *)
                    audio_stream_blocks->MakeArray();
                parentPtr->apackets = audio_stream_blocks->Length() - 1;

                parentPtr->audio_sample_list = (MPEGSampleInfo *)
                    audio_stream_samples->MakeArray();
                parentPtr->aframes = audio_stream_samples->Length();

                parentPtr->audio_first_header = audioParser->getHeader();
            }
        }

        parentPtr->video_sample_list = (MPEGSampleInfo *)
            video_stream_samples->MakeArray();
        parentPtr->frames = video_stream_samples->Length();


	} catch(const MyError&) {
		if (hWndStatus) {
			EnableWindow(GetParent(hWndStatus), TRUE);
			DestroyWindow(hWndStatus);
			hWndStatus = NULL;
		}
		throw;
	}

	if (hWndStatus) {
		EnableWindow(GetParent(hWndStatus), TRUE);
		DestroyWindow(hWndStatus);
		hWndStatus = NULL;
	}
}

INT_PTR CALLBACK MPEGFileParser::ParseDialogProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	MPEGFileParser *thisPtr = (MPEGFileParser *)GetWindowLongPtr(hDlg, DWLP_USER);

	switch(uMsg) {

	case WM_INITDIALOG:
		{
			TCHAR buf[64];

			SetWindowLongPtr(hDlg, DWLP_USER, lParam);
			thisPtr = (MPEGFileParser *)lParam;

			SendMessage(hDlg, WM_SETTEXT, 0, (LPARAM)"MPEG Import Filter");

	// Important note:
	// The value of "fIsMPEG2" depends on "fInterleaved."
	// If fInterleaved is true, fIsMPEG2 refers to the system layer
	// If fInterleaved is false, fIsMPEG2 refers to the video stream

			if (thisPtr->fIsVCD) {
				wsprintf(buf, "Parsing VideoCD stream");
			} else {
				if (thisPtr->fInterleaved) {
					wsprintf(buf, "Parsing interleaved MPEG-%ld file",
						(thisPtr->fIsMPEG2)? 2: 1);
				} else {
					wsprintf(buf, "Parsing MPEG-%ld video file",
						(thisPtr->fIsMPEG2)? 2: 1);
				}
			}
			SetDlgItemText(hDlg, IDC_STATIC_MESSAGE, (LPCTSTR)buf);

			SendMessage(GetDlgItem(hDlg, IDC_PROGRESS), PBM_SETRANGE, 0, MAKELPARAM(0, 16384));
			SetTimer(hDlg, 1, 250, (TIMERPROC)NULL);

			EnableWindow(GetParent(hDlg), FALSE);
			ShowWindow(hDlg, IsIconic(g_hWnd) ? SW_SHOWMINNOACTIVE : SW_SHOW);
			return TRUE;
		}

	case WM_TIMER:
		{
			TCHAR buf[64];
			__int64 file_cpos = thisPtr->Tell();

			SendMessage(GetDlgItem(hDlg, IDC_PROGRESS), PBM_SETPOS,
				(WPARAM)((file_cpos*16384) / thisPtr->file_len), 0);

			if (thisPtr->fIsVCD) {
				wsprintf(buf, "%ld of %ld sectors",
					(long)(file_cpos/2352),
					(long)(thisPtr->file_len/2352));
			} else {
				wsprintf(buf, "%ldK of %ldK",
					(long)(file_cpos>>10),
					(long)((thisPtr->file_len+1023)>>10));
			}

			SendDlgItemMessage(hDlg, IDC_CURRENT_VALUE, WM_SETTEXT, 0, (LPARAM)buf);
		}
		return TRUE;

	case WM_COMMAND:
		if (LOWORD(wParam) == IDCANCEL)
			thisPtr->fAbort = true;
		return TRUE;
	}

	return FALSE;
}


//////////////////////////////////////////////////////////////////////////
//
//
//							InputFileMPEG
//
//
//////////////////////////////////////////////////////////////////////////

const char InputFileMPEG::szME[]="MPEG Import Filter";

InputFileMPEG::InputFileMPEG() {
    // clear variables

    video_packet_buffer = NULL;
    video_packet_list = NULL;
    video_sample_list = NULL;
    audio_packet_list = NULL;
    audio_sample_list = NULL;
    audio_first_header = 0;

	video_field_map = NULL;

	firstPTS = 0;
	afirstPTS = 0;

    audioSrc = NULL;
    videoSrc = NULL;

	iChosenAudio = 0;
    iDecodeMode = 0;
    fAcceptPartial = true;

	bForceFILM = false;
	fAllowMatrix = false;

	seq_ext = 0;

    pFastRead = NULL;

    last_packet[0] = last_packet[1] = 0;
	CPUTest();
}

InputFile *CreateInputFileMPEG() {
    return new InputFileMPEG();
}

void InputFileMPEG::Init(const wchar_t *szFile) {
    VDLogAppMessage(kVDLogMarker, kVDST_Mpeg, kVDM_OpeningFile, 1, &szFile);

	// input options (if any) already exist at this point
	const int p = VDPreferencesGetMPEG2Prefs1();
	this->fAllowDSC = !!(p & MPEG2_ALLOW_DSC);
	this->bForceFILM = !!(p & MPEG2_DISABLE_RFF);
	this->fAllowMatrix = !!(p & MPEG2_ALLOW_MATRIX);

    AddFilename(szFile);

    // see if we can open the file

    try {
        mFile.open(szFile, nsVDFile::kRead | nsVDFile::kDenyWrite | nsVDFile::kOpenExisting | nsVDFile::kRandomAccess);

        pFastRead = new FastReadStream(mFile.getRawHandle(), 24, 32768);

		// Create and invoke MPEGFileParser

        MPEGFileParser fileParser(this);
        fileParser.Parse(&mFile);

		// Voila; the MPEG is parsed.

        if (fileParser.fIsPartial && !fAcceptPartial)
            VDLogAppMessage(kVDLogWarning, kVDST_Mpeg, kVDM_Incomplete);

		if ((fileParser.fInterleaved && packets <= 0) || frames <= 0)
			throw MyError("%s: no video stream!", szME);

		// Copy over the info we need

        this->width = fileParser.videoParser->width;
        this->height = fileParser.videoParser->height;
        this->mFrameRate = fileParser.videoParser->mFrameRate;
		this->seq_ext = fileParser.videoParser->seq_ext;
        this->fInterleaved = fileParser.fInterleaved;
        this->progressive_sequence = fileParser.videoParser->progressive_sequence;
        this->aspect_ratio = fileParser.videoParser->aspect_ratio_info;

#ifndef STORE_TIME_STAMPS
		this->firstPTS = fileParser.vEarliestPTS;
		this->afirstPTS = fileParser.aEarliestPTS;
		if (firstPTS == 0x7FFFFFFFFFFFFFFF) firstPTS = 0;
		if (afirstPTS == 0x7FFFFFFFFFFFFFFF) afirstPTS = 0;
#endif

		// MPEG-2 aspect ratio?
		if (seq_ext) aspect_ratio += 16;

#ifdef REMEMBER_AUDIO
		// This is scary and I have no idea how safe it is...

		if (this->iChosenAudio == 0) {
			this->iChosenAudio = fileParser.substream;

			InputFileMPEGOptions ifmo;

			ifmo.opts.len = sizeof(ifmo.opts);
			ifmo.opts.fAcceptPartial = this->fAcceptPartial;
			ifmo.opts.iDecodeMode = this->iDecodeMode;
			ifmo.opts.iChosenAudio = this->iChosenAudio;

			if (g_pInputOpts) {
				g_pInputOpts->read((char *)&ifmo.opts);
			} else {
				g_pInputOpts = createOptions((char *)&ifmo.opts);
			}
		}
#else
		this->iChosenAudio = fileParser.substream;
#endif

		// MPEGFileParser gets destructed now

    } catch(const MyError&) {
        throw;
    }

    // initialize DubSource pointers

	switch (iChosenAudio & AUDIOTYPE_MASK) {
		case AUDIOTYPE_AC3:
            audioSrc = new AudioSourceAC3(this);
            break;
		case AUDIOTYPE_LPCM:
            audioSrc = new AudioSourcePCM(this);
            break;
		case AUDIOTYPE_MPEG:
            audioSrc = new AudioSourceMPEG(this);
            break;
		default:
			iChosenAudio = 0;
			audioSrc = NULL;
			break;
	}

    if (audioSrc) audioSrc->init();

    if (!PostInit()) throw MyMemoryError();

	vdrefptr<IVDMPEGDecoder> pVideoDecoder(CreateVDMPEGDecoder());
    videoSrc = new VideoSourceMPEG(this, pVideoDecoder);

	if (videoSrc == NULL) throw MyMemoryError();

	if (!(videoSrc->init())) {
		throw MyError("VideoSourceMPEG: Couldn't initialize stream");
	}
}

bool InputFileMPEG::PostInit() {
    long i, j, count;
	long largest;

	// If there are field pictures, they must come in pairs.
	// Combine matching pairs into frame pictures and adjust
	// the total number of frames.

	// This must be done BEFORE reordering pictures

	i = 0;
	while (i < (frames - 1)) {
		MPEGSampleInfo *m1 = &(video_sample_list[i]);
		MPEGSampleInfo *m2 = &(video_sample_list[i+1]);

		if ((m1->picture_structure == TOP_FIELD &&
			m2->picture_structure == BOTTOM_FIELD) ||
			(m1->picture_structure == BOTTOM_FIELD &&
			m2->picture_structure == TOP_FIELD)) {

			if (m1->frame_type == m2->frame_type ||
				(m1->frame_type == MPEG_FRAME_TYPE_I &&
				m2->frame_type == MPEG_FRAME_TYPE_P)) {

				// Combine them, and remove the second field
				m1->size += m2->size;
				m1->fields = 2;

				for (j = i+1; j < (frames - 1); ++j) {
					video_sample_list[j] = video_sample_list[j+1];
				}
				frames--;
			}
		}
		++i;
	}

	// Assign Presentation Time Stamps

	// Note:  We know in which packet each access unit begins,
	// regardless if its start code straddles a packet boundary.

	// Note:  Do this BEFORE reordering the pictures, because
	// the reordering will disrupt the linear progression of
	// the "stream_pos" field.

#ifdef STORE_TIME_STAMPS
	if (fInterleaved) {

		count = 0;
		for (i = 0; i < packets; ++i) {
			if (video_packet_list[i].PTS != -1) {
				for (j = count; j < frames; ++j) {
					// Check BOTH boundaries!
					if (video_sample_list[j].stream_pos >= video_packet_list[i].stream_pos) {
						if ((i + 1) >= packets) goto AssignItDammit;
						if (video_sample_list[j].stream_pos < video_packet_list[i+1].stream_pos) {
AssignItDammit:				video_sample_list[j].PTS = video_packet_list[i].PTS;
							count = j + 1;
							break;
						}
					}
				}
			}
		}

		if (audioSrc) {
			// Assign Audio Presentation Time Stamps
			count = 0;
			for (i = 0; i < apackets; ++i) {
				if (audio_packet_list[i].PTS != -1) {
					for (j = count; j < aframes; ++j) {
						if (audio_sample_list[j].stream_pos >= audio_packet_list[i].stream_pos) {
							if ((i + 1) >= apackets) goto AssignIt2;
							if (audio_sample_list[j].stream_pos < audio_packet_list[i+1].stream_pos) {
AssignIt2:						audio_sample_list[j].PTS = audio_packet_list[i].PTS;
								count = j + 1;
								break;
							}
						}
					}
				}
			}
		}
	}
#endif

    // Now bubble-sort the frames into display order,
	for (i = 1; i < frames; ++i) {
		if (video_sample_list[i].frame_type == MPEG_FRAME_TYPE_B) {
			MPEGSampleInfo m = video_sample_list[i];
			video_sample_list[i] = video_sample_list[i-1];
			video_sample_list[i-1] = m;
		}
	}

	// In VTS_01_4, the broken sequence is BBPBBIBBPBBP
	// After reordering, it will be:
	// i = 1:  BBPBBIBBPBBP
	// i = 2:  BBPBBIBBPBBP
	// i = 3:  BBBPBIBBPBBP
	// i = 4:  BBBBPIBBPBBP
	// i = 5:  BBBBPIBBPBBP
	// i = 6:  BBBBPBIBPBBP
	// i = 7:  BBBBPBBIPBBP
	// i = 8:  BBBBPBBIPBBP
	// i = 9:  BBBBPBBIBPBP
	// i = 10: BBBBPBBIBBPP


	// To be fully decodable, the resulting sequence must
	// begin with an I-frame, and mustn't end with a B-frame

	// Remove leading P/B frames
	for (i = 0; i < frames; ++i) {
		if (video_sample_list[i].frame_type == MPEG_FRAME_TYPE_I) {
#ifdef STORE_TIME_STAMPS
			if (!fInterleaved) break;
			if (video_sample_list[i].PTS != -1)
#endif
			break;
		}
	}
	// "i" is now a count of how many frames precede the first I-frame
	if (i > 0) {
		if (i >= frames) return false;	// No I-frames at all?
		frames -= i;
		for (j = 0; j < frames; ++j) {
			video_sample_list[j] = video_sample_list[j+i];
		}
	}

	// Remove trailing B frames
	while (video_sample_list[frames - 1].frame_type == MPEG_FRAME_TYPE_B)
		frames--;

#ifdef STORE_TIME_STAMPS
	if (audioSrc) {

		// To ensure an accurate skew calculation, remove
		// any audio frames which precede the first audio PTS.
		for (i = 0; i < aframes; ++i) {
			if (audio_sample_list[i].PTS != -1) break;
		}

		// "i" is now a count of how many audio frames to remove
		if (i > 0) {
			if (i >= aframes) return false;	// No audio at all?
			aframes -= i;
			for (j = 0; j < aframes; ++j) {
				audio_sample_list[j] = audio_sample_list[j+i];
			}
		}
	}
#endif

	// Generate field map
	// While we're at it, determine the largest frame

	// "Force FILM" is a work in progress.  For now,
	// just emulate the 1.5.10 behavior and ignore RFF

	if (bForceFILM) {
		int accum;

		// For every 5 fields, we drop 1 field

		if (!(video_field_map = new long[frames]))
			return false;

		count = 0;
		largest = 0;
		accum = 0;

		/*
		i = 0;
		while (i < frames) {
			MPEGSampleInfo *m1, *m2;

			m1 = &(video_sample_list[i]);
			if (m1->size > largest) largest = m1->size;

			if (i >= (frames - 1)) {
				video_field_map[count++] = i * 2;
				break;
			}

			m2 = &(video_sample_list[i + 1]);
			if (m2->size > largest) largest = m2->size;

			if (	(m1->bitflags & PICTUREFLAG_RFF)
				&&	(m2->bitflags & PICTUREFLAG_RFF)) {

				// Two consecutive frames with RFF
				VDDEBUG("Two consecutive frames with RFF");

			} else if (	(m1->bitflags & PICTUREFLAG_RFF) == 0
					&&	(m2->bitflags & PICTUREFLAG_RFF) == 0) {

				// Two consecutive frames without RFF
				VDDEBUG("Two consecutive frames (%i) without RFF\n", i);

				if (++accum > 1) {
					// Only add the first frame
					video_field_map[count++] = i * 2 + 1;
					accum -= 2;
				} else {
					// Add both frames
					video_field_map[count++] = i * 2;
					video_field_map[count++] = (i + 1) * 2;
				}

			} else {

				// One with RFF, one without RFF

				video_field_map[count++] = i * 2;
				video_field_map[count++] = (i + 1) * 2;
			}

			i += 2;
		}

		fields = count;
		mFrameRate = VDFraction(24000, 1001);
		*/


		for (i = 0; i < frames; ++i) {
			const MPEGSampleInfo *m1 = &(video_sample_list[i]);
			if (m1->size > largest) largest = m1->size;
			video_field_map[i] = i * 2;
			count += m1->fields;
		}

		fields = frames;

        // Detect 3:2 pulldown and adjust the frame rate
		// Either:	fields = frames * 2 (perfect NTSC)
		// Or:		fields = frames * 2.5 (perfect FILM)

        if (mFrameRate.getHi() == 30000 && mFrameRate.getLo() == 1001) {
            if (count >= (long)(frames * 2.3)) {
                mFrameRate = VDFraction(24000, 1001);
			}
        }

	} else {

		count = 0;
		j = 0;	// cadence

		for (i = 0; i < frames; ++i) {
			const MPEGSampleInfo *m1 = &(video_sample_list[i]);

			// this is overly complicated
			if (	(m1->bitflags & PICTUREFLAG_RFF)
				&&	(m1->picture_structure == FRAME_PICTURE)
				&&	(m1->bitflags & PICTUREFLAG_PF)) {

				if (this->progressive_sequence) {
					if (m1->bitflags & PICTUREFLAG_TFF) {
						count += 3;	// 3 frames
					} else {
						count += 2;	// 2 frames
					}
				} else {
					++count;		// 3 fields
					j = 1 - j;		// flip the cadence
					if (j == 0) ++count;
				}

			} else {
				++count;			// 1 frame, or 2 fields
			}
		}

		if (!(video_field_map = new long[count]))
			return false;

		// Now load the map
		fields = count;
		count = 0;

		j = 0;	// cadence
		largest = 0;

		for (i = 0; i < frames; ++i) {
			const MPEGSampleInfo *m1 = &(video_sample_list[i]);

			if (m1->size > largest) largest = m1->size;

			if (	(m1->bitflags & PICTUREFLAG_RFF)
				&&	(m1->picture_structure == FRAME_PICTURE)
				&&	(m1->bitflags & PICTUREFLAG_PF)) {

				// Add 1 progressive frame
				video_field_map[count++] = i * 2 + j;

				if (this->progressive_sequence) {
					// 2 progressive frames
					video_field_map[count++] = i * 2 + j;
					if (m1->bitflags & PICTUREFLAG_TFF) {
						// 3 progressive frames
						video_field_map[count++] = i * 2 + j;
					}
				} else {
					j = 1 - j;	// flip cadence
					if (j == 0) {
						// Add a duplicate frame
						video_field_map[count++] = i * 2 + j;
					}
				}

			} else {
				// 1 progressive frame, or 2 field pictures
				video_field_map[count++] = i * 2 + j;
			}
		}
	}

#ifdef STORE_TIME_STAMPS
	// Find the earliest video and audio timestamps

	if (fInterleaved) {
		for (i = 0; i < frames; ++i) {
			if (video_sample_list[i].PTS != -1) {
				firstPTS = video_sample_list[i].PTS;
				break;
			}
		}
		afirstPTS = firstPTS;

		if (audioSrc != NULL) {
			for (i = 0; i < aframes; ++i) {
				if (audio_sample_list[i].PTS != -1) {
					afirstPTS = audio_sample_list[i].PTS;
					break;
				}
			}
		}
	}
#endif

    // Set video_packet_buffer to size of the largest frame
    count = (largest + 15) & -15;

	largest_framesize = count;
    if (video_packet_buffer = new char[count * 2]) return true;
    return false;
}

InputFileMPEG::~InputFileMPEG() {
    //_RPT0(0,"Destructor called\n");

	videoSrc = NULL;
	audioSrc = NULL;

    delete [] video_packet_buffer;
    delete video_packet_list;
    delete video_sample_list;
    delete audio_packet_list;
    delete audio_sample_list;

	delete [] video_field_map;

    if (pFastRead)
        delete pFastRead;
}

//////////////////////

void InputFileMPEG::ReadStream(void *buffer, __int64 pos, long len, bool fAudio) {
    int pkt = 0;
    long delta;
    char *ptr = (char *)buffer;
    MPEGPacketInfo *packet_list = fAudio ? audio_packet_list : video_packet_list;
    int pkts = fAudio ? apackets : packets;

//  _RPT2(0,"Stream read request: pos %ld, len %ld\n", pos, len);

    if (!fInterleaved) {

        if (pFastRead)
            pFastRead->Read(0, pos, buffer, len);
        else {
            mFile.seek(pos);
            mFile.read(buffer, len);
        }
        return;
    }

    // find the packet containing the data start using a binary search

    do {
        int l = 0;
        int r = pkts-1;

        // Check current packet

        pkt = last_packet[fAudio?1:0];

        if (pkt>=0 && pkt<pkts) {
            if (pos < packet_list[pkt].stream_pos)
                r = pkt-1;
            else if (pos < packet_list[pkt+1].stream_pos)
                break;
            else if (pkt+1 < pkts && pos < packet_list[pkt+2].stream_pos) {
                ++pkt;
                break;
            } else
                l = pkt+2;
        }

        for(;;) {
            if (l > r) {
                pkt = l;
                break;
            }

            pkt = (l+r)/2;

            if (pos < packet_list[pkt].stream_pos)
                r = pkt-1;
            else if (pos >= packet_list[pkt+1].stream_pos)
                l = pkt+1;
            else
                break;
        }

        if (pkt<0 || pkt >= pkts) throw MyError("MPEG Internal error: Invalid stream read position (%ld)", pos);

    } while(false);

    // offset into packet for this stream position:
    delta = (long)(pos - packet_list[pkt].stream_pos);

    // read data from packets

    while(len) {
        if (pkt >= pkts)
			throw MyError("Attempt to read past end of stream (pos %ld)", pos);

        long tc = (long)((packet_list[pkt+1].stream_pos - packet_list[pkt].stream_pos) - delta);

        if (tc>len) tc=len;

//              if (!fAudio)
//                      _RPT3(0,"Reading %ld bytes at %08lx+%ld\n", tc, video_packet_list[pkt].file_pos,delta);

        if (pFastRead) {
            pFastRead->Read(fAudio ? 1 : 0, packet_list[pkt].file_pos + delta, ptr, tc);
        } else {
            mFile.seek(packet_list[pkt].file_pos + delta);
            mFile.read(ptr, tc);
        }
        len -= tc;
        ptr += tc;
        ++pkt;
        delta = 0;
    }

    last_packet[fAudio?1:0] = pkt-1;
}

///////////////////////////////////////////////////////////////////////////

// Here is the sequence of events in project.cpp:
// Either "createOptions" or "promptForOptions" is called
// (or maybe neither and g_pInputOpts isn't created)
//
// If "extended open" is checked, "promptForOptions" is called
// If "pInputOptions" is non-NULL, "createOptions" is called
//
// Immediately after this, "setOptions" is called.

// Finally, the input file "Init" method is called.

void InputFileMPEG::setOptions(InputFileOptions *_ifo) {
    InputFileMPEGOptions *ifo = (InputFileMPEGOptions *)_ifo;

    iDecodeMode = ifo->opts.iDecodeMode;

    if (iDecodeMode & InputFileMPEGOptions::DECODE_NO_P)
        iDecodeMode |= InputFileMPEGOptions::DECODE_NO_B;

#ifdef REMEMBER_AUDIO
	iChosenAudio = ifo->opts.iChosenAudio;
#endif
}

InputFileOptions *InputFileMPEG::createOptions(const char *buf) {
    InputFileMPEGOptions *ifo = new InputFileMPEGOptions();

    if (!ifo) throw MyMemoryError();

    if (!ifo->read(buf)) {
        delete ifo;
        return NULL;
    }

    return ifo;
}

InputFileOptions *InputFileMPEG::promptForOptions(HWND hwnd) {
    InputFileMPEGOptions *ifo = new InputFileMPEGOptions();

    if (!ifo) throw MyMemoryError();

    DialogBoxParam(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_EXTOPENOPTS_MPEG), hwnd, InputFileMPEGOptions::SetupDlgProc, (LPARAM)ifo);

    return ifo;
}

///////////////////////////////////////////////////////////////////////////

typedef struct MyFileInfo {
    InputFileMPEG *thisPtr;

    volatile HWND hWndAbort;
    UINT statTimer;

    long lFrames;
    __int64 lTotalSize;
    long lFrameCnt[3];
    long lFrameMinSize[3];
    long lFrameMaxSize[3];
    long lFrameTotalSize[3];
    long lAudioSize;
    long lAudioAvgBitrate;
    const char *lpszAudioMode;

} MyFileInfo;

void InputFileMPEG::_InfoDlgThread(void *pvInfo) {
    MyFileInfo *pInfo = (MyFileInfo *)pvInfo;
    InputFileMPEG *thisPtr = pInfo->thisPtr;
    long i;
    VideoSourceMPEG *vSrc = (VideoSourceMPEG *)&*pInfo->thisPtr->videoSrc;
    AudioSourceMPEG *aSrc = (AudioSourceMPEG *)&*pInfo->thisPtr->audioSrc;
    MPEGSampleInfo *msi;

	const VDPosition videoFrameStart	= vSrc->getStart();
	const VDPosition videoFrameEnd		= pInfo->thisPtr->frames;

    for(i=0; i<3; i++)
        pInfo->lFrameMinSize[i] = 0x7FFFFFFF;

    msi = thisPtr->video_sample_list;
    for(i = (long)videoFrameStart; i < (long)videoFrameEnd; ++i) {
        int iFrameType = msi->frame_type;

        if (iFrameType) {
            long lSize = msi->size;
            --iFrameType;

            ++pInfo->lFrameCnt[iFrameType];
            pInfo->lFrameTotalSize[iFrameType] += lSize;

            if (lSize < pInfo->lFrameMinSize[iFrameType])
                pInfo->lFrameMinSize[iFrameType] = lSize;

            if (lSize > pInfo->lFrameMaxSize[iFrameType])
                pInfo->lFrameMaxSize[iFrameType] = lSize;

            pInfo->lTotalSize += lSize;

        }
        ++pInfo->lFrames;

        ++msi;
        if (pInfo->hWndAbort) {
            SendMessage(pInfo->hWndAbort, WM_USER+256, 0, 0);
            return;
        }
    }

    ///////////////////////////////////////////////////////////////////////

    if (aSrc) {
        static const char *szModes[8] =
			{"stereo", "joint stereo", "dual channel", "mono",
			 "3 channels", "4 channels", "5 channels", "5.1"};

        double lTotalBitrate = 0;

        switch (thisPtr->iChosenAudio & AUDIOTYPE_MASK) {

        case AUDIOTYPE_MPEG: {
            bool fAudioMono = false;
            bool fAudioMixedMode = false;
            unsigned int audio_first_mode = MPEGAudioHeader(thisPtr->audio_first_header).GetMode();
            fAudioMono = !MPEGAudioHeader(thisPtr->audio_first_header).IsStereo();
            msi = thisPtr->audio_sample_list;

            for(i = thisPtr->aframes; i > 0; i--) {
                MPEGAudioHeader fAudioHeader(msi->header);

                if (!fAudioMixedMode) {
                    if (fAudioHeader.GetMode() != audio_first_mode)
                        fAudioMixedMode = true;
                }

                lTotalBitrate += fAudioHeader.GetBitrateKbps();
                pInfo->lAudioSize += msi->size;
                ++msi;

                if (pInfo->hWndAbort) {
                    SendMessage(pInfo->hWndAbort, WM_USER+256, 0, 0);
                    return;
                }
            }

            if (thisPtr->aframes)
                lTotalBitrate /= thisPtr->aframes;
            else lTotalBitrate = 0;
            pInfo->lAudioAvgBitrate = (long)lTotalBitrate;

            if (fAudioMixedMode) {
                if (fAudioMono)
                    pInfo->lpszAudioMode = "mixed mode";
                else
                    pInfo->lpszAudioMode = "mixed stereo";
            } else
                pInfo->lpszAudioMode = szModes[(thisPtr->audio_first_header>>30) & 3];

            }
            break;

        case AUDIOTYPE_AC3: {
			WAVEFORMATEX *wfex = aSrc->getWaveFormat();
			pInfo->lAudioAvgBitrate = wfex->nAvgBytesPerSec / 125;
			if (wfex->nChannels > 2) {
				pInfo->lpszAudioMode = szModes[wfex->nChannels+1];
			} else if (wfex->nChannels > 1) {
				pInfo->lpszAudioMode = szModes[0];
			} else {
				pInfo->lpszAudioMode = szModes[3];
			}
			pInfo->lAudioSize = (long)(aSrc->getEnd() * wfex->nBlockAlign);
            } break;

        case AUDIOTYPE_LPCM: {
            WAVEFORMATEX *wfex = aSrc->getWaveFormat();
            pInfo->lAudioAvgBitrate = wfex->nAvgBytesPerSec / 125;
            pInfo->lAudioSize = thisPtr->aframes * 1536*2 * wfex->nChannels;
            pInfo->lpszAudioMode = szModes[(wfex->nChannels > 1)? 0: 3];
            } break;
        }
    }

    pInfo->hWndAbort = (HWND)1;
}

INT_PTR APIENTRY InputFileMPEG::_InfoDlgProc( HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam) {
    MyFileInfo *pInfo = (MyFileInfo *)GetWindowLongPtr(hDlg, DWLP_USER);
    InputFileMPEG *thisPtr;

	static const char *sz_aspect_ratios[32] = {
		// MPEG-1 SAR
		"SAR = 0 (forbidden)",
		"SAR = 1 (1.0) 1:1 VGA",
		"SAR = 2 (0.6735)",
		"SAR = 3 (0.7031) 16:9 PAL",
		"SAR = 4 (0.7615)",
		"SAR = 5 (0.8055)",
		"SAR = 6 (0.8437) 16:9 NTSC",
		"SAR = 7 (0.8935)",
		"SAR = 8 (0.9157) 4:3 PAL",
		"SAR = 9 (0.9815)",
		"SAR = 10 (1.0255)",
		"SAR = 11 (1.0695)",
		"SAR = 12 (1.095) 4:3 NTSC",
		"SAR = 13 (1.1575)",
		"SAR = 14 (1.2015)",
		"SAR = 15 (reserved)",
		// MPEG-2 SAR/DAR
		"DAR = 0 (forbidden)",
		"SAR = 1 (1.0)",
		"DAR = 2 (4:3)",
		"DAR = 3 (16:9)",
		"DAR = 4 (2.21:1)",
		"DAR = 5 (reserved)",
		"DAR = 6 (reserved)",
		"DAR = 7 (reserved)",
		"DAR = 8 (reserved)",
		"DAR = 9 (reserved)",
		"DAR = 10 (reserved)",
		"DAR = 11 (reserved)",
		"DAR = 12 (reserved)",
		"DAR = 13 (reserved)",
		"DAR = 14 (reserved)",
		"DAR = 15 (reserved)"
	};

    if (pInfo)
        thisPtr = pInfo->thisPtr;

    switch (message)
    {
        case WM_INITDIALOG:
            SetWindowLongPtr(hDlg, DWLP_USER, lParam);
            pInfo = (MyFileInfo *)lParam;
            thisPtr = pInfo->thisPtr;

            if (thisPtr->videoSrc != NULL) {
				char *s;
				char buf[128];
				int i;

				sprintf(buf, "MPEG-%ld Information",
					thisPtr->seq_ext? 2: 1);
				SetWindowText(hDlg, buf);

				sprintf(buf, "%dx%d, %.3f fps (%ld s)",
							thisPtr->width,
							thisPtr->height,
							thisPtr->videoSrc->getRate().asDouble(),
							VDRoundToLong(1000000.0 / thisPtr->videoSrc->getRate().asDouble()));
				SetDlgItemText(hDlg, IDC_VIDEO_FORMAT, buf);

				i = thisPtr->aspect_ratio & 31;
				//if (thisPtr->seq_ext) i += 16;
				SetDlgItemText(hDlg, IDC_ASPECT_RATIO, sz_aspect_ratios[i]);

				s = buf + sprintf(buf, "%lu (", (unsigned)thisPtr->videoSrc->getLength());
				ticks_to_str(s, VDRoundToLong(1000.0*thisPtr->videoSrc->getLength()/thisPtr->videoSrc->getRate().asDouble()));
				strcat(s,")");
				SetDlgItemText(hDlg, IDC_VIDEO_NUMFRAMES, buf);
			}
            if (thisPtr->audioSrc != NULL) {
                WAVEFORMATEX *fmt = thisPtr->audioSrc->getWaveFormat();
                char buf[128];

                sprintf(buf, "%ldHz, %s", fmt->nSamplesPerSec, fmt->nChannels>1 ? "Stereo" : "Mono");
                SetDlgItemText(hDlg, IDC_AUDIO_FORMAT, buf);

                sprintf(buf, "%ld", thisPtr->aframes);
                SetDlgItemText(hDlg, IDC_AUDIO_NUMFRAMES, buf);
            }
            _beginthread(_InfoDlgThread, 10000, pInfo);

            pInfo->statTimer = SetTimer(hDlg, 1, 250, NULL);

            return (TRUE);

        case WM_COMMAND:
            if (LOWORD(wParam) == IDOK || LOWORD(wParam) == IDCANCEL)
            {
                if (pInfo->hWndAbort == (HWND)1)
                    EndDialog(hDlg, TRUE);
                return TRUE;
            }
            break;

        case WM_DESTROY:
            if (pInfo->statTimer) KillTimer(hDlg, pInfo->statTimer);
            break;

        case WM_TIMER:
            {
                char buf[128];

                sprintf(buf, "%ld", pInfo->lFrames);
                SetDlgItemText(hDlg, IDC_VIDEO_NUMKEYFRAMES, buf);

                sprintf(buf, "%ld / %ld / %ld", pInfo->lFrameCnt[0], pInfo->lFrameCnt[1], pInfo->lFrameCnt[2]);
                SetDlgItemText(hDlg, IDC_VIDEO_FRAMETYPECNT, buf);

                UINT uiCtlIds[]={ IDC_VIDEO_IFRAMES, IDC_VIDEO_PFRAMES, IDC_VIDEO_BFRAMES };

                for(int i=0; i<3; i++) {

                    if (pInfo->lFrameCnt[i])
                        sprintf(buf, "%ld / %ld / %ld (%ldK)"
                                    ,pInfo->lFrameMinSize[i]
                                    ,pInfo->lFrameTotalSize[i]/pInfo->lFrameCnt[i]
                                    ,pInfo->lFrameMaxSize[i]
                                    ,(pInfo->lFrameTotalSize[i]+1023)>>10);
                    else
                        sprintf(buf,"(no %c-frames)", "IPB"[i]);

                    SetDlgItemText(hDlg, uiCtlIds[i], buf);
                }

                if (pInfo->lFrames) {
                    long lKBitsPerSec, lKBytesPerSec;

	// This won't work for movies with RFF, because we don't know
	// the display rate of the encoded frames (it may even vary).
	// However, we do know the playback duration, so divide that
	// by the number of coded frames to get an average frame rate.

	// bits * (frames per sec) / frames = bits per sec
	// (a / b) / c  =  a / (b * c)
	// a / (b / c)  =  (a * c) / b

//					double d = (pInfo->lTotalSize *
//						thisPtr->videoSrc->getRate().asDouble())
//						/ (double)pInfo->lFrames;

					// First, calculate the AVERAGE frame rate:
					double d = ((double)pInfo->lFrames *
						thisPtr->videoSrc->getRate().asDouble())
						/ (double)thisPtr->videoSrc->getLength();

					d = (d * pInfo->lTotalSize) / (double)pInfo->lFrames;

                    lKBytesPerSec = VDRoundToLong(d / 1024.0);
                    lKBitsPerSec = VDRoundToLong(d * 0.008);

                    sprintf(buf, "%ld Kbps (%ldKB/s)", lKBitsPerSec, lKBytesPerSec);
                    SetDlgItemText(hDlg, IDC_VIDEO_AVGBITRATE, buf);
                }

                if (pInfo->lpszAudioMode && thisPtr->audioSrc) {
                    static const char *szLayers[]={"I","II","III"};
                    WAVEFORMATEX *fmt = thisPtr->audioSrc->getWaveFormat();

                    switch (thisPtr->iChosenAudio & AUDIOTYPE_MASK) {
                        case AUDIOTYPE_MPEG:
                            sprintf(buf, "%ldKHz %s, %ldKbps layer %s",
                              fmt->nSamplesPerSec/1000,
                              pInfo->lpszAudioMode,
                              pInfo->lAudioAvgBitrate,
                              //szLayers[((thisPtr->audio_first_header>>17)&3)^3]
                              szLayers[((thisPtr->audio_first_header>>9)&3)^3]
                            );
                            break;

                        case AUDIOTYPE_AC3:
                            sprintf(buf, "%ldKHz %s, %ldKbps AC-3",
                              fmt->nSamplesPerSec/1000,
                              pInfo->lpszAudioMode,
                              pInfo->lAudioAvgBitrate
                            );
                            break;

                        case AUDIOTYPE_LPCM:
                            sprintf(buf, "%ldKHz %s, %ldKbps LPCM",
                              fmt->nSamplesPerSec/1000,
                              pInfo->lpszAudioMode,
                              pInfo->lAudioAvgBitrate
                            );
                            break;
                    }
                    SetDlgItemText(hDlg, IDC_AUDIO_FORMAT, buf);

                    sprintf(buf, "%ldK", (pInfo->lAudioSize + 1023) / 1024);
                    SetDlgItemText(hDlg, IDC_AUDIO_SIZE, buf);

					sprintf(buf, "%ld ms", VDRoundToLong((thisPtr->afirstPTS - thisPtr->firstPTS) / 90.0));
					SetDlgItemText(hDlg, IDC_AUDIO_SKEW, buf);
                }
            }

            /////////

            if (pInfo->hWndAbort) {
                KillTimer(hDlg, pInfo->statTimer);
                return TRUE;
            }

            break;

        case WM_USER+256:
            EndDialog(hDlg, TRUE);
            break;
    }
    return FALSE;
}

void InputFileMPEG::InfoDialog(HWND hwndParent) {
    MyFileInfo mai;

    memset(&mai, 0, sizeof mai);
    mai.thisPtr = this;

    DialogBoxParam(g_hInst, MAKEINTRESOURCE(IDD_MPEG_INFO), hwndParent, _InfoDlgProc, (LPARAM)&mai);
}

/////////////////////////////////////////////////////////////////////

class VDInputDriverMPEG : public vdrefcounted<IVDInputDriver> {
public:
	const wchar_t *GetSignatureName() { return L"MPEG input driver (internal)"; }

	int GetDefaultPriority() { return 0; }

	uint32 GetFlags() { return kF_Video | kF_Audio; }

	const wchar_t *GetFilenamePattern() {
		return L"MPEG video/systems stream (*.mpg,*.mpeg,*.mpv,*.m1v,*.m2v,*.m2p,*.vob,*.dat)\0*.mpg;*.mpeg;*.mpv;*.m1v;*.m2v;*.m2p;*.vob;*.dat\0";
	}

	bool DetectByFilename(const wchar_t *pszFilename) {
		if (pszFilename != NULL) {
			int len = 0;
			while (pszFilename[len] != 0) len++;

			if (len >= 4) {
				const wchar_t *ext = &pszFilename[len-4];
				if (!wcsncmp(ext, L".mpg", 4)) return true;
				if (!wcsncmp(ext, L".mpv", 4)) return true;
				if (!wcsncmp(ext, L".m1v", 4)) return true;
				if (!wcsncmp(ext, L".m2v", 4)) return true;
				if (!wcsncmp(ext, L".m2p", 4)) return true;
				if (!wcsncmp(ext, L".vob", 4)) return true;
				if (len >= 5) {
					const wchar_t *ext = &pszFilename[len-5];
					if (!wcsncmp(ext, L".mpeg", 5)) return true;
				}
			}
		}
		return false;
	}

	int DetectBySignature(const void *pHeader, sint32 nHeaderSize, const void *pFooter, sint32 nFooterSize, sint64 nFileSize) {
		if (nHeaderSize >= 12) {

			// I think this is how it works:
			//	we return  1 if we're certain it is
			//	we return  0 if we think it could be
			//	we return -1 if we're certain it's not

			const uint32 *src = (uint32 *)pHeader;

			// Check for VCD DAT
			if (src[0] == 'FFIR' && src[1] >= nHeaderSize && src[2] == 'AXDC') return 1;

			// Check for MPEG System/Video
			if (src[0] == 0xBA010000 || src[0] == 0xB3010000) return 1;

			nHeaderSize -= 3;
			if (nHeaderSize > 70000) nHeaderSize = 70000;

			// Scan for broken MPEG System/Video
			while (nHeaderSize > 0) {
				if (*src == 0xBA010000) break;
				if (*src == 0xB3010000) break;
				src = (uint32 *)((uint8 *)src + 1);
				--nHeaderSize;
			}

            if (nHeaderSize) return 0;
		}

		return -1;
	}

	InputFile *CreateInputFile(uint32 flags) {
		return new InputFileMPEG;
	}
};

extern IVDInputDriver *VDCreateInputDriverMPEG() { return new VDInputDriverMPEG; }
